"""A module to test `Process` and related classes."""

import pytest
import requests
import pandas as pd
from pandas.testing import assert_frame_equal
from unittest import mock
from fero.process import Process, Stage, Tag, DataRequestError
from io import BytesIO


@pytest.fixture
def process_fixture(process_data, patched_fero_client):
    """Get a sample `Process` object."""
    return Process(patched_fero_client, process_data)


@pytest.fixture
def stage_fixture(process_fixture, process_stages):
    """Get a sample list of `Stage` objects."""
    return [
        Stage(process_fixture, process_fixture._client, sdata)
        for sdata in process_stages["stages"]
    ]


@pytest.fixture
def tags_fixture(process_fixture, process_tags):
    """Get a sample list of `Tag` objects."""
    return [
        Tag(process_fixture, process_fixture._client, tdata)
        for tdata in process_tags["tags"]
    ]


@pytest.fixture
def process_with_loaded_data(process_fixture, stage_fixture, tags_fixture):
    """Get a sample `Process` object with loaded stages and tags."""
    with mock.patch.object(
        Process, "stages", new_callable=mock.PropertyMock
    ) as mock_stages:
        with mock.patch.object(
            Process, "tags", new_callable=mock.PropertyMock
        ) as mock_tags:
            mock_stages.return_value = stage_fixture
            mock_tags.return_value = tags_fixture
            yield process_fixture


def test_process_get_stages(
    process_fixture, patched_fero_client, process_stages, stage_fixture
):
    """Test that a process can get stages."""
    patched_fero_client.get.return_value = process_stages
    assert process_fixture.stages == stage_fixture
    # test caching
    assert process_fixture.stages == stage_fixture

    patched_fero_client.get.assert_called_once_with(
        f"/api/processes/{process_fixture.api_id}/stages/"
    )


def test_process_get_tags(
    process_fixture, patched_fero_client, process_tags, tags_fixture
):
    """Test that a process can get tags."""
    patched_fero_client.get.return_value = process_tags
    assert process_fixture.tags == tags_fixture
    # test caching
    assert process_fixture.tags == tags_fixture

    patched_fero_client.get.assert_called_once_with(
        f"/api/processes/{process_fixture.api_id}/tags/"
    )


def test_get_tag_stage_with_tag(process_with_loaded_data):
    """Test getting a stage by providing a tag name."""
    assert process_with_loaded_data.get_tag_stage("s2_factor1").id == 100


def test_get_tag_stage_with_tag_use_tag_obj(process_with_loaded_data, tags_fixture):
    """Test getting a stage by providing a tag."""
    tag = next(tag for tag in tags_fixture if tag.name == "s2_factor1")
    assert process_with_loaded_data.get_tag_stage(tag).id == 100


def test_get_tag_stage_no_tag(process_with_loaded_data):
    """Test getting a stage that doesn't exist."""
    assert process_with_loaded_data.get_tag_stage("not_a_factor") is None


def test_stages_by_kpis_with_tag_name(process_with_loaded_data, stage_fixture):
    """Test getting stages by providing a kpi name."""
    assert (
        process_with_loaded_data.get_stages_by_kpis(["s2_factor1"]) == stage_fixture[:2]
    )


def test_stages_by_kpis_with_tag_multiple(process_with_loaded_data, stage_fixture):
    """Test getting stages by providing multiple kpi name."""
    assert (
        process_with_loaded_data.get_stages_by_kpis(["s2_factor1", "s1_factor1"])
        == stage_fixture[:2]
    )


def test_stages_by_kpis_with_tag_use_tag_obj(
    process_with_loaded_data, tags_fixture, stage_fixture
):
    """Test getting a stages by providing a tag kpi."""
    tag = next(tag for tag in tags_fixture if tag.name == "s2_factor1")
    assert process_with_loaded_data.get_stages_by_kpis([tag]) == stage_fixture[:2]


def test_stages_by_kpis_no_tag(process_with_loaded_data):
    """Test getting a stages if the kpi doesn't exist."""
    assert process_with_loaded_data.get_stages_by_kpis(["not_a_factor"]) == []


def test_get_data_works(process_with_loaded_data, patched_fero_client):
    """Test that get data makes the correct calls and loads a file."""
    patched_fero_client.post.return_value = {
        "request_id": "abcd-1234",
        "data": {"download_url": "test.com", "success": True, "message": "success"},
        "complete": True,
    }

    test_df = pd.DataFrame({"s1_factor1": [1, 2, 3], "s3_factor1": ["x", "y", "z"]})
    test_file = BytesIO()
    test_df.to_parquet(test_file)
    test_file.seek(0)
    mock_request = mock.MagicMock()
    mock_request.iter_content.return_value = [test_file.read()]
    with mock.patch("requests.get", return_value=mock_request):
        df = process_with_loaded_data.get_data(["s1_factor1", "s3_factor1"])

    test_file.close()
    assert_frame_equal(test_df, df)

    patched_fero_client.post.assert_has_calls(
        [
            mock.call(
                f"/api/processes/{process_with_loaded_data.api_id}/download_process_data/",
                {"request_data": {"tags": ["s1_factor1", "s3_factor1"], "kpis": None}},
            ),
            mock.call(
                f"/api/processes/{process_with_loaded_data.api_id}/download_process_data/",
                {"request_id": "abcd-1234"},
            ),
        ]
    )


def test_get_data_continuous_no_kpi(process_with_loaded_data):
    """Test that get data raises an error for a continuous process and with no kpis."""
    process_with_loaded_data._data["process_type"] = "C"

    with pytest.raises(DataRequestError) as exc:
        process_with_loaded_data.get_data(["s1_factor1", "s3_factor1"])
        assert (
            exc.message == "A kpi must be specified to download continuous process data"
        )


def test_get_data_unsuccesful(process_with_loaded_data, patched_fero_client):
    """Test that a fero error is raised if the server indicates an unsuccesful request."""
    patched_fero_client.post.return_value = {
        "request_id": "abcd-1234",
        "data": {
            "download_url": None,
            "success": False,
            "message": "something bad happened!",
        },
        "complete": True,
    }

    with pytest.raises(DataRequestError) as exc:
        process_with_loaded_data.get_data(["s1_factor1", "s3_factor1"])
        assert exc.message == "something bad happened!"

    patched_fero_client.post.assert_has_calls(
        [
            mock.call(
                f"/api/processes/{process_with_loaded_data.api_id}/download_process_data/",
                {"request_data": {"tags": ["s1_factor1", "s3_factor1"], "kpis": None}},
            ),
            mock.call(
                f"/api/processes/{process_with_loaded_data.api_id}/download_process_data/",
                {"request_id": "abcd-1234"},
            ),
        ]
    )


def test_get_data_raises_bad_status(
    monkeypatch, process_with_loaded_data, patched_fero_client
):
    """Test that an exception is raised for non-2xx status codes."""

    class MockResponse:
        status_code = 404

        def raise_for_status(self):
            if 400 <= self.status_code < 600:
                raise requests.exceptions.HTTPError(response=self)

    def mock_get(*args, **kwargs):
        return MockResponse()

    def mock_raise_for_status(*args, **kwargs):
        mock_response.raise_for_status()

    monkeypatch.setattr(requests, "get", mock_get)
    monkeypatch.setattr(requests.Response, "raise_for_status", mock_raise_for_status)
    mock_response = MockResponse()

    with pytest.raises(requests.exceptions.HTTPError) as exc:
        process_with_loaded_data.get_data(["s1_factor1", "s3_factor1"])
        assert exc.value.response.status_code in range(400, 500)



"""This module defines classes and functions used in multiple places throughout the fero client."""

import fero
import time
import requests
from typing import Any, Callable, TypeVar
from marshmallow import Schema

from fero.exceptions import FeroError


class FeroObject:
    """
    A base class for fero-related objects.

    Data related to the object, formatted as a `schema_class`,
    is stored in the `_data` attribute and accessible through a normal
    `__getattr__` call (like a dictionary).
    """

    schema_class: Schema = None

    def __init__(self, client: "fero.Fero", data: dict):
        """
        Create a new `FeroObject` using provided `data`.

        :param client: The fero client object that gives access to this object
        :param data: A dictionary holding values for the fields of a `schema_class`
        """
        self._client = client
        schema = self.schema_class()
        self._data = schema.load(data)

    def __getattr__(self, name: str) -> Any:
        """
        Access an entry in this object's schema.

        :param name: the entry in the schema to access
        :return: the data stored in this entry of the schema
        """
        return self._data.get(name)


class FeroTimeoutError(FeroError):
    """An error raised when an api request times out."""

    def __init__(self, last_response: requests.Response):
        """
        Create a `FeroTimeoutError` citing the most recent response `last_response`.

        :param last_response: the last response from the api before timing out
        """
        self.last_response = last_response
        super().__init__("Request exceeded specified polling timeout")


D = TypeVar("D")


def poll(
    func: Callable[[], D], test: Callable[[D], bool], interval=0.5, timeout=300
) -> D:
    """Poll the api until a test condition is met.

    :param func: function to do the polling request
    :param test: Function to test whether the polling should end
    :param interval: interval to poll at, defaults to 0.5
    :param timeout: timeout value to stop polling, defaults to 300
    :raises FeroTimeoutError: raised after `timeout` seconds if no value passes `test`
    :return: the object to be returned from the function
    """
    data = func()
    start_time = time.time()
    while not test(data):
        if time.time() - start_time > timeout:
            raise FeroTimeoutError(data)
        time.sleep(interval)
        data = func()

    return data



"""This module defines an error to be used for fero-specific exceptions."""


class FeroError(Exception):
    """A base error for fero-specific exceptions."""

    def __init__(self, *args):
        """
        Create a `FeroError` with arbitrary positional arguments.

        :param args: an argument list of arbitrary length.
            The first argument, if present, is the message of the error.
        """
        self.message = args[0] if len(args) > 0 else None
        super().__init__(*args)



"""This module holds classes related to a fero analysis."""

import time
import fero
import uuid
import io
import datetime
from fero import FeroError
import pandas as pd
from marshmallow import (
    Schema,
    fields,
    validate,
    validates_schema,
    ValidationError,
    EXCLUDE,
)
from typing import Any, Union, List, Optional, IO
from .common import FeroObject


VALID_GOALS = ["minimize", "maximize"]
FERO_COST_FUNCTION = "FERO_COST_FUNCTION"
V1_RESULT_SUFFIXES = ["low", "mid", "high"]
V2_RESULT_SUFFIXES = ["low90", "low50", "mid", "high50", "high90"]
BULK_PREDICTION_TYPE = "M"


class AnalysisSchema(Schema):
    """A schema to store data related to a fero analysis."""

    class Meta:
        """
        Specify that unknown fields included on this schema should be excluded.

        See https://marshmallow.readthedocs.io/en/stable/quickstart.html#handling-unknown-fields.
        """

        unknown = EXCLUDE

    url = fields.String(required=True)

    predictions_url = fields.String(required=True)
    revisions_url = fields.String(required=True)

    created_by = fields.Dict(required=True)

    data_source_name = fields.String(required=False)
    process_name = fields.String(required=False)

    data_source_deleted = fields.Boolean()
    process_deleted = fields.Boolean()

    latest_revision = fields.Integer()

    latest_revision_model = fields.UUID(required=True)

    latest_revision_model_state = fields.String(required=True)

    latest_trained_revision = fields.Integer(required=False, allow_none=True)

    latest_trained_revision_model = fields.UUID(allow_none=True)

    latest_completed_model = fields.UUID(allow_none=True)
    # We should eventually make this required again when we update the server with defaults
    latest_completed_model_score = fields.Integer(required=False, allow_none=True)

    latest_completed_model_score_qualifier = fields.String(
        required=False, allow_none=True
    )

    latest_completed_model_modified = fields.DateTime(required=False, allow_none=True)

    latest_completed_model_schema = fields.Dict(default=dict, missing=dict)

    schema_overrides = fields.Dict(default=dict, missing=dict)

    display_options = fields.Dict(allow_none=True, default=dict, missing=dict)
    ac_name = fields.String(required=True)

    uuid = fields.UUID(required=True)
    name = fields.String(required=True)
    data_source = fields.UUID(required=True, allow_none=True)
    process = fields.UUID(required=False, allow_none=True)
    created = fields.DateTime(require=True)
    modified = fields.DateTime(require=True)
    blueprint_name = fields.String(required=True)


class RevisionSchema(Schema):
    """A schema to store data related to a fero analysis revision."""

    class Meta:
        """
        Specify that unknown fields included on this schema should be excluded.

        See https://marshmallow.readthedocs.io/en/stable/quickstart.html#handling-unknown-fields.
        """

        unknown = EXCLUDE

    version = fields.Integer()
    configured_blueprint = fields.Dict()


class FactorSchema(Schema):
    """A schema to store data related to a factor linked to a fero analysis optimization."""

    name = fields.String(required=True)
    min = fields.Number(required=False, default=None)
    max = fields.Number(required=False, default=None)
    cost = fields.Number(required=False)

    @validates_schema
    def relative_min_and_max(self, data: dict, **kwargs):
        """Validate that either both min and max or none, or they are both defined, and min < max.

        :raises ValidationError: if min >= max or only one is defined.
        """
        min = data.get("min", None)
        max = data.get("max", None)
        if min is None and max is None:
            return
        elif max is None:
            raise ValidationError(
                {"max": ["A value for 'max' must be provided when 'min' is present."]}
            )
        elif min is None:
            raise ValidationError(
                {"min": ["A value for 'min' must be provided when 'max' is present."]}
            )
        elif min >= max:
            raise ValidationError(
                {"min": ["The value of 'min' must be less than the value of 'max'."]}
            )


class CostSchema(FactorSchema):
    """A schema to store data related to a cost linked to a fero analysis optimization."""

    cost = fields.Float(required=True)

    @validates_schema
    def numeric_factors(self, data: dict, **kwargs):
        """Validate that the cost function factors, being numeric, include min and max parameters.

        :raises ValidationError: if no min or max parameter is defined
        """
        min = data.get("min", None)
        max = data.get("max", None)
        if min is None or max is None:
            raise ValidationError(
                {"factor": ["Cost factors must include both a min and max."]}
            )


class BaseGoalSchema(Schema):
    """A base schema to store data related to the goal of a fero analysis optimization."""

    goal = fields.String(validate=validate.OneOf(VALID_GOALS))


class StandardOptimizeGoal(BaseGoalSchema):
    """A schema to store data related to a goal to min/max some tag in a fero analysis optimization."""

    factor = fields.Nested(FactorSchema(), required=True)

    @validates_schema
    def numeric_goal(self, data: dict, **kwargs):
        """Validate that the goal, being numeric, includes a min and max parameter.

        :raises ValidationError: if no min or max parameter is defined
        """
        if (
            data["factor"].get("min", None) is None
            or data["factor"].get("max", None) is None
        ):
            raise ValidationError(
                {"factor": ["Optimization goal factor must include a min and max."]}
            )


class CostOptimizeGoal(BaseGoalSchema):
    """A schema to store data related to a goal to minimize cost in a fero analysis optimization."""

    type = fields.String(validate=validate.OneOf(["COST"]), required=True)
    cost_function = fields.Nested(CostSchema, many=True, required=True)

    @validates_schema
    def max_functions(self, data: dict, **kwargs):
        """Validate that there are no more than three cost functions in the `data`.

        :raises ValidationError: when there are more than three cost functions specified in `data`
        """
        if len(data["cost_function"]) > 3:
            raise ValidationError(
                {
                    "cost_function": [
                        "No more than three factors allowed in the cost function."
                    ]
                }
            )


class Prediction:
    """Represents the results of a prediction submitted to Fero.

    Predictions are run asynchrounously by Fero and data will only be available if the `complete` property
    is true.  Attempting to access the data for the prediction before it is complete will result in a `FeroError`
    being raised.
    """

    def __init__(self, client: "fero.Fero", prediction_result_id: str):
        """
        Create a `Prediction` using the ID from a set of prediction results.

        :param client: a fero client to interface with the API
        :param prediction_result_id: the ID tied to the prediction's result
        """
        self._client = client
        self.result_id = prediction_result_id
        self._data_cache = None
        self._complete = False
        self._result_id = None

    def __repr__(self) -> str:
        """Represent the `Prediction` by its prediction result ID."""
        return f"<Prediction request_id={self.result_id}>"

    def __getattr__(self, name: str) -> Any:
        """
        Access an entry in this object's schema.

        :param name: the entry in the schema to access
        :return: the data stored in this entry of the schema
        """
        return self._data.get(name)

    __str__ = __repr__

    @property
    def _data(self):
        """Get the data related to this prediction result if it is available."""
        # check if it's complete which also pulls data
        if not self.complete:
            return None

        return self._data_cache

    @property
    def status(self):
        """Check the status of the prediction if it is available."""
        if not self.complete:
            return None
        if "result_data" in self._data and "status" in self._data["result_data"]:
            return self._data["result_data"]["status"]
        return None

    @property
    def complete(self):
        """Check if the Prediction is complete."""
        if not self._complete:
            self._data_cache = self._client.get(
                f"/api/prediction_results/{self.result_id}/"
            )
            self._complete = self._data_cache["state"] == "C"

        return self._complete

    def get_results(self, format="dataframe") -> Union[pd.DataFrame, List[dict]]:
        """Return the prediction results of the prediction.

        By default this will be a pandas DataFrame, but specifying `format="record"`
        will instead return a list of dictionaries where each key specifies
        a factor and the value is the prediction.

        :param format: The format to return the result as, defaults to "dataframe"
        :type format: str, optional
        :raises FeroError: Raised if the prediction is not yet complete or is completed and failed.
        :return: The results of the prediction
        :rtype: Union[pd.DataFrame, List[dict]]
        """
        if not self.complete:
            raise FeroError("Prediction is not complete.")
        if self.status != "SUCCESS":
            raise FeroError(
                f"Prediction failed with the following message: {self._data['result_data']['message']}"
            )
        if self.prediction_type == BULK_PREDICTION_TYPE:
            data_url = self._data["result_data"]["data"]["download_url"]
            data = self._client.get_preauthenticated(data_url)
            return pd.DataFrame(**data)
        else:
            data = self._data["result_data"]["data"]["values"]
            if format == "records":
                return [
                    {col: val for col, val in zip(data["columns"], row)}
                    for row in data["data"]
                ]
            else:
                return pd.DataFrame(
                    data["data"], columns=data["columns"], index=data["index"]
                )


class Analysis(FeroObject):
    """An object for interacting with a specific Analysis on Fero.

    The Analysis is the primary way to access a model associated with a data set.
    Once an analysis is created it can be used to perform various actions such as making a prediction
    based on provided data or optimizing certain values based on the Fero model of the data.
    """

    _presentation_data_cache: Union[dict, None] = None
    _reg_factors: Union[List[dict], None] = None
    _reg_targets: Union[List[dict], None] = None
    _factor_names: Union[List[str], None] = None
    _target_names: Union[List[str], None] = None

    schema_class = AnalysisSchema

    def __repr__(self) -> str:
        """Represent the `Analysis` by its name."""
        return f"<Analysis name={self.name}>"

    __str__ = __repr__

    @property
    def _regression_factors(self) -> Optional[List[dict]]:
        """Get data on the factors of the Analysis."""
        if self._reg_factors is None:
            self._parse_regression_data()

        return self._reg_factors

    @property
    def _regression_targets(self) -> Optional[List[dict]]:
        """Get data on the targets of the Analysis."""
        if self._reg_targets is None:
            self._parse_regression_data()

        return self._reg_targets

    @property
    def factor_names(self) -> Optional[List[str]]:
        """Get the names of the factors associated with the Analysis."""
        if self._factor_names is None:
            self._parse_regression_data()

        return self._factor_names

    @property
    def target_names(self) -> Optional[List[str]]:
        """Get the names of the targets of the Analysis."""
        if self._target_names is None:
            self._parse_regression_data()

        return self._target_names

    @property
    def _presentation_data(self) -> Optional[dict]:
        """Get the presentation data of this analysis."""
        if self._presentation_data_cache is None:
            self._get_presentation_data()

        return self._presentation_data_cache

    @property
    def _schema(self):
        """Get the schema for the process related to this Analysis."""
        if self._schema_cache is None:
            v1_schema = self._data.get("latest_completed_model_schema", None)
            if (v1_schema is None or len(v1_schema) == 0) and self.process is not None:
                v1_schema = self._client.get(
                    f"/api/processes/{self.process}/v1_interpreted_schema/"
                )
            elif (
                v1_schema is None or len(v1_schema) == 0
            ) and self.data_source is not None:
                v1_schema = self._client.get(
                    f"/api/data_sources/{self.data_source}/latest_interpreted_schema/"
                )
            self._schema_cache = v1_schema

        return self._schema_cache

    @staticmethod
    def _make_col_name(col_name: str, cols: List[str]) -> str:
        """Mangle duplicate columns."""
        c = 0
        og_col_name = col_name
        while col_name in cols:
            col_name = f"{og_col_name}.{c}"
            c += 1

        return col_name

    def is_retraining(self) -> bool:
        """Check if an analysis is currently being retrained.

        :return: True if the analysis is in the process of being retrained, false otherwise.
        """
        self._refresh_analysis()
        training = self.latest_revision_model_state == "T"
        latest_revision = self.latest_revision
        if not training:
            self._refresh_analysis()
            # requery if a new version somehow got created when we refreshed.  Do this until they match
            if self.latest_revision != latest_revision:
                return self.is_retraining()

        return training

    def revise(self):
        """Trigger an analysis revision which causes the analysis to be retrained with any new data or changes in the process configurations.

        This method will not revise the analysis if a new analysis is currently being trained.  It currently only supports revising an analysis with the previously selected options.

        For more granular configuration of a revision please use the Fero Labs website.
        """
        # don't revise if currently training
        if self.is_retraining():
            return

        latest_revision = self._client.get(
            f"/api/analyses/{self.uuid}/revisions/{self.latest_revision}/"
        )
        latest_revision = RevisionSchema().load(latest_revision)
        latest_blueprint = latest_revision["configured_blueprint"]

        self._client.post(
            f"/api/analyses/{self.uuid}/revisions/",
            {
                "description": f"Revision created via fero client on {datetime.datetime.now().isoformat()}",
                "configured_blueprint": latest_blueprint,
            },
        )
        self._refresh_analysis()

    def _refresh_analysis(self):
        """Reload the latest analysis data from the server."""
        data = self._client.get(f"/api/analyses/{self.uuid}/")
        schema = AnalysisSchema()
        self._data = schema.load(data)

    def _upload_file(self, fp: IO, file_tag: str, prediction_type: str) -> str:
        """Upload a single file to the uploaded files location and returns ID of workspace to check."""
        inbox_response = self._client.post(
            f"/api/analyses/{self.uuid}/workspaces/inbox_url/",
            {"file_tag": file_tag, "prediction_type": prediction_type},
        )

        workspace_id = inbox_response.pop("workspace_id")
        self._client.upload_file(inbox_response, file_tag, fp)
        return workspace_id

    def _parse_regression_data(self):
        """Get and parse the regression simulator object from presentation data."""
        reg_data = next(
            d
            for d in self._presentation_data["data"]
            if d["id"] == "regression_simulator"
        )
        self._reg_factors = reg_data["content"]["factors"]
        self._reg_targets = reg_data["content"]["targets"]
        self._factor_names = [f["factor"] for f in self._reg_factors]
        self._target_names = [t["name"] for t in self._reg_targets]

    def _flatten_result(self, result: dict, prediction_row: dict) -> dict:
        """Flatten nested results returned by the api into a single dict and combine it with the provided data."""
        flat_result = {}
        cols = prediction_row.keys()

        for target, values in result["data"].items():
            suffix_list = (
                V2_RESULT_SUFFIXES if result["version"] == 2 else V1_RESULT_SUFFIXES
            )
            flat_result.update(
                {
                    self._make_col_name(f"{target}_{suffix}", cols): values["value"][
                        suffix
                    ][0]
                    for suffix in suffix_list
                }
            )
        flat_result.update(prediction_row)
        return flat_result

    def _get_presentation_data(self):
        """Retrieve the presentation data of this analysis revision from the api."""
        self._presentation_data_cache = self._client.get(
            f"/api/revision_models/{self.latest_trained_revision_model}/presentation_data/"
        )

    def has_trained_model(self) -> bool:
        """Check whether this analysis has a trained model associated with it.

        :return: True if there is a model, False otherwise
        """
        return self.latest_completed_model is not None

    def make_prediction(
        self, prediction_data: Union[pd.DataFrame, List[dict]]
    ) -> Union[pd.DataFrame, List[dict]]:
        """Make a prediction from the provided data using the most recent trained model for the analysis.

        This method is optimized for analyses that support fast, bulk prediction.

        `make_prediction` takes either a data frame or list of dictionaries of values that will be sent to Fero
        to make a prediction of what the targets of the Analysis will be. The results are returned as either a dataframe
        or list of dictionaries with both the original prediction data and the predicted targets in each row or dict.
        Prediction data should not contain any missing values. Each target has a `high`, `low`, and `mid` value and
        these are added to the target variable name with an `_`.

        :param prediction_data:  Either a data frame or list of dictionaries specifying values to be used in the model.
        :type prediction_data: Union[pd.DataFrame, List[dict]]
        :raises FeroError: Raised if no model has been trained or the server returns an error message
        :return: A data frame or list of dictionaries depending on how the function was called
        :rtype: Union[pd.DataFrame, List[dict]]
        """
        if not self.has_trained_model:
            raise FeroError("No model has been trained on this analysis.")

        is_dict_list = isinstance(prediction_data, list)

        prediction_df = (
            pd.DataFrame(prediction_data) if is_dict_list else prediction_data
        )

        data_file = io.StringIO()
        prediction_df.to_json(data_file, orient="split")
        data_file.seek(0)
        upload_identifier = str(uuid.uuid4())
        workspace_id = self._upload_file(
            data_file, upload_identifier, BULK_PREDICTION_TYPE
        )
        prediction = self._poll_workspace_for_prediction(workspace_id)
        if prediction.status != "SUCCESS":
            raise FeroError(
                prediction.result_data.get(
                    "message", "The prediction failed for unknown reasons"
                )
            )
        output = prediction.get_results()
        return list(output.T.to_dict().values()) if is_dict_list else output

    def _get_factor_dtype(self, factor_name: str) -> Union[str, None]:
        """Return the dtype of a factor or `None` if the factor isn't found."""
        try:
            goal_data = next(
                f for f in self._regression_factors if f["factor"] == factor_name
            )
            return f"factor_{goal_data['dtype']}"
        except StopIteration:
            return None

    def _get_target_dtype(self, target_name: str) -> Union[str, None]:
        """Return the dtype of a target or `None` if the target isn't found."""
        try:
            # Only support real and integer targets
            guess = next(
                f["guess"]
                for f in self._schema["columns"]
                if f["name"] == target_name and f["guess"] in ["real", "integer"]
            )
            return "target_float" if guess == "real" else "target_int"
        except StopIteration:
            return None

    def _verify_standard_goal(self, goal: dict):
        """Verify the goal config relative to the analysis."""
        goal_name = goal["factor"]["name"]

        # The goal label must be a target or factor
        if goal_name not in self.target_names + self.factor_names:
            raise FeroError(f'"{goal_name}" is not a target or factor for this model.')

        # If this is a factor makes sure it's not a float
        if goal_name not in self.target_names and self._get_factor_dtype(
            goal_name
        ) not in ["factor_float", "factor_int"]:
            raise FeroError("The data type of the goal must be float or integer.")

    def _verify_cost_goal(self, goal: dict):
        """Verify that a cost goal is correct."""
        for c in goal["cost_function"]:
            name = c["name"]
            factor_type = self._get_factor_dtype(name)
            # Implicitly find missing factors
            if factor_type is not None and factor_type not in [
                "factor_float",
                "factor_int",
            ]:
                raise FeroError(
                    "The data type of all cost function factors must be float or integer."
                )
            else:
                target_type = self._get_target_dtype(name)
                if target_type is None:
                    raise FeroError(
                        f'"{name}" is not a factor or target in this model.'
                    )
                # Implicitly find missing factors
                if target_type not in [
                    "target_float",
                    "target_int",
                ]:
                    raise FeroError(
                        "The data type of all cost function targets must be float or integer."
                    )

    def _verify_constraints(self, constraints: List[dict], **kwargs):
        """Verify provided constraints are in the analysis."""
        use_adaptive = kwargs.get("use_adaptive", False)
        for constraint in constraints:
            constraint_name = constraint["name"]
            constraint_type = (
                self._get_factor_dtype(constraint_name)
                if constraint_name in self.factor_names
                else self._get_target_dtype(constraint_name)
                if constraint_name in self.target_names
                else None
            )
            if constraint_type is None:
                raise FeroError(
                    f'Constraint "{constraint_name}" is not part of this model.'
                )
            elif constraint_type == "factor_category" and (
                constraint.get("min", None) is not None
                or constraint.get("max", None) is not None
            ):
                raise FeroError(
                    f'Categorical factor "{constraint_name}" should not define a min or max value.'
                )
            elif use_adaptive and constraint_type == "factor_category":
                raise FeroError(
                    "Categorical factors are not yet supported for adaptive optimizations."
                )
            elif constraint_type in [
                "factor_int",
                "factor_float",
                "target_int",
                "target_float",
            ] and (
                constraint.get("min", None) is None
                or constraint.get("max", None) is None
            ):
                raise FeroError(
                    f'Numeric constraint "{constraint_name}" requires a min and max value.'
                )

    def _cross_verify_optimization(self, goal: dict, constraints: List[dict], **kwargs):
        """Verify the config has the correct combined targets, costs and factors."""
        is_cost = goal.get("type", None) == "COST"

        cost_factors = []
        goal_factor = []
        target_factor = []
        constraint_factors = []
        constraint_targets = []
        use_adaptive = kwargs.get("use_adaptive", False)

        if len(constraints) < 1:
            raise FeroError("At least one constraint must be specified.")
        if is_cost:
            for factor in goal["cost_function"]:
                if factor["name"] in self.factor_names:
                    cost_factors.append(factor["name"])
                else:
                    target_factor.append(factor["name"])
        else:
            if goal["factor"]["name"] in self.factor_names:
                goal_factor.append(goal["goal"])
            else:
                target_factor.append(goal["goal"])

        for constraint in constraints:
            if constraint["name"] in self.factor_names:
                constraint_factors.append(constraint["name"])
            else:
                constraint_targets.append(constraint["name"])

        if len(target_factor + constraint_targets) < 1:
            raise FeroError("No Targets specified")

        if (
            len(constraint_factors + cost_factors + goal_factor) > 3
            and not use_adaptive
        ):
            raise FeroError(
                "A maximum of three factors can be specified in an optimization."
            )
        elif use_adaptive and len(constraint_factors + cost_factors + goal_factor) > 5:
            raise FeroError(
                "A maximum of five factors can be specified in an adaptive optimization."
            )

    def _verify_fixed_factors(self, fixed_factors: dict):
        """Check that the provided fixed factors are in the analysis."""
        all_columns = self.target_names + self.factor_names
        for key in fixed_factors.keys():
            if key not in all_columns:
                raise FeroError(f'"{key}" is not part of this model.')

    def _get_basis_values(self):
        """Get median fixed values from presentation data."""
        reg_data = next(
            d
            for d in self._presentation_data["data"]
            if d["id"] == "regression_simulator"
        )
        # yuck massive ugly json
        base_values = {f["factor"]: f["median"] for f in reg_data["content"]["factors"]}

        target_values = {
            t["name"]: t["default"]["mid"][0] for t in reg_data["content"]["targets"]
        }

        base_values.update(target_values)
        return base_values

    def _build_optimize_request(
        self,
        name,
        goal,
        constraints,
        fixed_factors,
        include_confidence_intervals,
        **kwargs,
    ) -> dict:
        """Format the content for an optimization request."""
        is_cost = goal.get("type", None) == "COST"
        ci_value = "include" if include_confidence_intervals else "exclude"
        opt_request = {
            "name": name,
            "description": "",
            "prediction_type": "O",
        }

        input_data = {
            "kind": "OptimizationRequest",
            "version": 1,
            "objectives": [
                {
                    "factor": FERO_COST_FUNCTION if is_cost else goal["factor"]["name"],
                    "goal": goal["goal"],
                }
            ],
            "basisSpecifiedColumns": [],  # These appear to just be, uhm, here
            "linearFunctionDefinitions": {},
            "useAdaptiveGrid": kwargs.get("use_adaptive", False),
        }
        basis_values = self._get_basis_values()
        basis_values.update(fixed_factors)
        input_data["basisValues"] = basis_values

        factor_bounds = []
        for c in constraints:
            if c["name"] in self.factor_names:
                dtype = self._get_factor_dtype(c["name"])
                factor_bounds.append(
                    {
                        "factor": c["name"],
                        "lowerBound": c.get("min", None),
                        "upperBound": c.get("max", None),
                        "dtype": dtype,
                    }
                    if dtype != "factor_category"
                    else {"factor": c["name"], "dtype": dtype}
                )
        target_bounds = [
            {
                "factor": c["name"],
                "lowerBound": c.get("min", None),
                "upperBound": c.get("max", None),
                "dtype": self._get_target_dtype(c["name"]),
                "confidenceInterval": ci_value,
            }
            for c in constraints
            if c["name"] in self.target_names
        ]

        bounds = factor_bounds + target_bounds

        if is_cost:
            cost_lower_bound = 0
            cost_upper_bound = 0

            for factor in goal["cost_function"]:
                cost_lower_bound += factor["min"] * factor["cost"]
                cost_upper_bound += factor["max"] * factor["cost"]

            for c in goal["cost_function"]:
                factor_is_target = c["name"] in self.target_names
                bound = {
                    "factor": c["name"],
                    "lowerBound": c["min"],
                    "upperBound": c["max"],
                    "dtype": f"{'target' if factor_is_target else 'factor'}_float",
                }
                if factor_is_target:
                    bound["confidenceInterval"] = ci_value
                bounds.append(bound)

            bounds.append(
                {
                    "factor": FERO_COST_FUNCTION,
                    "lowerBound": cost_lower_bound,
                    "upperBound": cost_upper_bound,
                    "dtype": "function",
                }
            )

            input_data["linearFunctionDefinitions"] = {
                FERO_COST_FUNCTION: {
                    c["name"]: c["cost"] for c in goal["cost_function"]
                }
            }

        else:
            goal_name = goal["factor"]["name"]
            goal_is_target = goal_name in self.target_names
            dtype = (
                self._get_target_dtype(goal_name)
                if goal_is_target
                else self._get_factor_dtype(goal_name)
            )
            goal_bound = {
                "factor": goal["factor"]["name"],
                "lowerBound": goal["factor"]["min"],
                "upperBound": goal["factor"]["max"],
                "dtype": dtype
                if dtype is not None
                else f"{'target' if goal_is_target else 'factor'}_float",
            }
            if goal_is_target:
                goal_bound["confidenceInterval"] = ci_value

            bounds.append(goal_bound)

        input_data["bounds"] = bounds
        opt_request["input_data"] = input_data

        return opt_request

    def _request_prediction(
        self, prediction_request: dict, synchronous: bool
    ) -> Prediction:
        """Make the prediction request and poll unitl complete if this request is synchronous."""
        response_data = self._client.post(
            f"/api/analyses/{str(self.uuid)}/workspaces/",
            {"request": prediction_request, "visible": False},
        )
        prediction = Prediction(
            self._client, response_data["latest_prediction"]["latest_results"]
        )

        # If synchronous block until the prediction is complete
        while synchronous and not prediction.complete:
            time.sleep(0.5)

        return prediction

    def _poll_workspace_for_prediction(self, workspace_id: str):
        """Poll workspace until it and a result exist."""
        workspace_url = f"/api/analyses/{str(self.uuid)}/workspaces/{workspace_id}"
        workspace_data = None
        while workspace_data is None:
            time.sleep(0.5)
            workspace_data = self._client.get(
                workspace_url,
                params={"prediction_type": BULK_PREDICTION_TYPE},
                allow_404=True,
            )

        prediction = Prediction(
            self._client, workspace_data["latest_prediction"]["latest_results"]
        )
        while not prediction.complete:
            time.sleep(0.5)

        return prediction

    def make_optimization(
        self,
        name: str,
        goal: dict,
        constraints: List[dict],
        fixed_factors: Optional[dict] = None,
        include_confidence_intervals: bool = True,
        synchronous: bool = True,
        **kwargs,
    ) -> Prediction:
        """Perform an optimization using the most recent model for the analysis.

        By default this function will block until the optimization is complete, however specifying `synchonous=False`
        will instead return a prediction object referencing the optimization being made. This prediction will not contain
        results until the `complete` property is true.

        The expected config input looks as follows:

        Example configuration for a standard (without cost) optimization::

            {
                "goal": "maximize",
                "factor": {"name": "factor1", "min": 5, "max": 10}
            }

        Example configuration for a cost optimization:
        Cost Goal Config
        ::

            {
                "type": "COST",
                "goal": "minimize"
                "cost_function": [
                    {"min": 5, "max": 10, "cost": 1000, "name": "factor1"},
                    {"min": 5, "max": 10, "cost": 500,  "name": "target2"}
                ]
            }

        The constraints configuration is a list of factors and their constraints::

            [
                {"name": "factor2",  "min": 10, "max": 10}
                {"name": "target1", "min": 100, "max": 500}
            ]

        :param name: Name for this optimization
        :type name: str
        :param goal: A dictionary describing the goal of the optimization
        :type goal: dict
        :param constraints: A dictionary describing the constraints of the optimization
        :type constraints: dict
        :param fixed_factors: Values of factors to stay fixed if not provided the mean values are used, defaults to None
        :type fixed_factors: dict, optional
        :param include_confidence_intervals: Whether the optimization should include the lower 5% and upper 95% prediction levels, defaults to True
        :type include_confidence_intervals: bool, optional
        :param synchronous: Whether the optimization should return only after being complete.  This can take a bit, defaults to True
        :type synchronous: bool, optional
        :return: The results of the optimization
        :rtype: Prediction
        """
        if self.blueprint_name == "fault":
            raise FeroError("Fault analysis optimization are not supported")

        cost_goal = "type" in goal
        if fixed_factors is None:
            fixed_factors = {}

        goal_schema = CostOptimizeGoal() if cost_goal else StandardOptimizeGoal()
        goal_validation = goal_schema.validate(goal)
        if goal_validation:
            raise FeroError(f"Error validating goal <f{str(goal_validation)}>")

        constraints_schema = FactorSchema(many=True)
        constraints_validation = constraints_schema.validate(constraints)
        if constraints_validation:
            raise FeroError(
                f"Error validating constraints <f{str(constraints_validation)}>"
            )

        if cost_goal:
            self._verify_cost_goal(goal)
        else:
            self._verify_standard_goal(goal)

        self._verify_constraints(constraints, **kwargs)
        self._cross_verify_optimization(goal, constraints, **kwargs)
        self._verify_fixed_factors(fixed_factors)

        optimize_request = self._build_optimize_request(
            name,
            goal,
            constraints,
            fixed_factors,
            include_confidence_intervals,
            **kwargs,
        )
        return self._request_prediction(optimize_request, synchronous)



"""A module to test the `Fero` class."""

from fero.datasource import DataSource
from fero.analysis import Analysis
from fero.asset import Asset
from fero.process import Process
import pytest
from unittest import mock
from pathlib import Path
from fero import Fero, FeroError


@pytest.fixture
def mock_response():
    """Create a mocked successful API response."""
    response = mock.MagicMock()
    response.json.return_value = {"token": "fakeToken"}
    response.content
    response.headers = {"content-type": "application/json"}
    response.status_code = 200

    return response


@pytest.fixture()
def patch_requests_post(mock_response):
    """Create a mocked successful POST response."""
    with mock.patch("fero.client.requests.post") as post_function:

        post_function.return_value = mock_response
        yield post_function


@pytest.fixture
def patch_requests_get(mock_response):
    """Create a mocked successful GET response."""
    with mock.patch("fero.client.requests.get") as get_function:

        get_function.return_value = mock_response
        yield get_function


@pytest.fixture
def mock_conf_path(monkeypatch, tmp_path_factory):
    """Create a mocked successful POST response."""
    path = tmp_path_factory.mktemp("home-")
    monkeypatch.setattr(Path, "home", mock.MagicMock(return_value=path))

    yield path


@pytest.fixture
def patch_fero_get():
    """Create a mocked GET call from the fero client."""
    with mock.patch.object(Fero, "get") as mock_fero_get:
        yield mock_fero_get


def test_fero_client_with_jwt():
    """Test that a client is created with a jwt provided."""
    client = Fero(fero_token="fakeToken")
    assert isinstance(client, Fero)
    assert client._fero_token == "fakeToken"


def test_fero_client_username_pass_provided(patch_requests_post):
    """Test that a client is created correctly with a provided username and password."""
    client = Fero(username="fero", password="pass")
    assert isinstance(client, Fero)

    patch_requests_post.assert_called_with(
        "https://app.ferolabs.com/api/token/auth/",
        json={"username": "fero", "password": "pass"},
        verify=True,
    )
    assert client._fero_token == "fakeToken"


def test_fero_client_env_token(monkeypatch):
    """Test that a token is found in the env."""
    monkeypatch.setenv("FERO_TOKEN", "fakeToken")

    client = Fero()
    assert isinstance(client, Fero)
    assert client._fero_token == "fakeToken"


def test_fero_client_token_file(mock_conf_path):
    """Test that a token is found in the .fero file."""
    with open(str(mock_conf_path / ".fero"), "w") as fero_file:
        fero_file.write("FERO_TOKEN=fakeToken\n")

    client = Fero()
    assert isinstance(client, Fero)
    assert client._fero_token == "fakeToken"


def test_fero_client_user_pass_file(patch_requests_post, mock_conf_path):
    """Test that username and password are found in the .fero file."""
    with open(str(mock_conf_path / ".fero"), "w") as fero_file:
        fero_file.write("FERO_USERNAME=fero\nFERO_PASSWORD=pass\n")

    client = Fero()
    assert isinstance(client, Fero)
    assert client._fero_token == "fakeToken"

    patch_requests_post.assert_called_with(
        "https://app.ferolabs.com/api/token/auth/",
        json={"username": "fero", "password": "pass"},
        verify=True,
    )


def test_fero_client_no_creds():
    """Test that an exception is raised if there is no way to get a token."""
    with pytest.raises(FeroError):
        Fero()


def test_fero_client_get_wrapper(patch_requests_get):
    """Test that get requests have the correct headers added."""
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    client.get("/some/url/", params={"n": "42"})
    patch_requests_get.assert_called_with(
        "http://test.com/some/url/",
        params={"n": "42"},
        headers={"Authorization": "JWT fakeToken"},
        verify=True,
    )


def test_fero_client_post_wrapper(patch_requests_post):
    """Test that post requests have the correct headers added."""
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    client.post("/some/url/", {"n": "42"})
    patch_requests_post.assert_called_with(
        "http://test.com/some/url/",
        json={"n": "42"},
        headers={"Authorization": "JWT fakeToken"},
        verify=True,
    )


def test_fero_get_raises_error_404(patch_requests_get, mock_response):
    """Test that a FeroError with the expected message is raised if the url isn't found."""
    mock_response.status_code = 404
    patch_requests_get.return_value = mock_response
    client = Fero(fero_token="fakeToken", hostname="http://test.com")

    with pytest.raises(FeroError) as err:
        client.get("/some/url/")
        assert err.message == "The requested resource was not found"


def test_fero_get_raises_error_not_authorized(patch_requests_get, mock_response):
    """Test that a FeroError with the expected message is raised if the request isn't authorized."""
    mock_response.status_code = 403
    patch_requests_get.return_value = mock_response
    client = Fero(fero_token="fakeToken", hostname="http://test.com")

    with pytest.raises(FeroError) as err:
        client.get("/some/url/")
        assert err.message == "The requested resource was not found."

    mock_response.status_code = 401
    with pytest.raises(FeroError) as err:
        client.get("/some/url/")
        assert err.message == "The requested resource was not found."


def test_fero_get_raises_error_400(patch_requests_get, mock_response):
    """Test that a `FeroError` with the expected message is raised if a 400 is returned."""
    mock_response.status_code = 400
    patch_requests_get.return_value = mock_response
    client = Fero(fero_token="fakeToken", hostname="http://test.com")

    with pytest.raises(FeroError) as err:
        client.get("/some/url/")
        assert (
            err.message == "There was an issue connecting to Fero. - Status Code: 400"
        )


def test_get_analysis_success(patch_fero_get, analysis_data):
    """Test that an analysis is returned by get_analysis."""
    patch_fero_get.return_value = analysis_data
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    analysis = client.get_analysis("some-uuid")
    assert isinstance(analysis, Analysis)
    assert analysis.name == analysis_data["name"]
    patch_fero_get.assert_called_with("/api/analyses/some-uuid/")


def test_search_analyses(patch_fero_get, analysis_data):
    """Test that the correct iterator of analyses is returned by `search_analyses`."""
    patch_fero_get.return_value = {"next": None, "results": [analysis_data]}
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    analyses = [a for a in client.search_analyses(analysis_data["name"])]
    assert len(analyses) == 1
    assert isinstance(analyses[0], Analysis)
    assert analyses[0].name == analysis_data["name"]
    patch_fero_get.assert_called_with(
        "/api/analyses/", params={"name": analysis_data["name"]}
    )


def test_search_analyses_paginated(patch_fero_get, analysis_data):
    """Test that a list analyses is returned by `search_analyses`."""
    patch_fero_get.side_effect = [
        {
            "next": "http://test.com/api/analyses/?token=1234",
            "results": [analysis_data],
        },
        {"next": None, "results": [analysis_data]},
    ]
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    analyses = [a for a in client.search_analyses(analysis_data["name"])]
    assert len(analyses) == 2
    assert isinstance(analyses[0], Analysis)
    assert analyses[0].name == analysis_data["name"]
    assert isinstance(analyses[1], Analysis)
    assert analyses[1].name == analysis_data["name"]
    patch_fero_get.assert_has_calls(
        [
            mock.call("/api/analyses/", params={"name": analysis_data["name"]}),
            mock.call("/api/analyses/?token=1234", params=None),
        ]
    )


def test_get_asset_success(patch_fero_get, asset_data):
    """Test that an asset is returned by `get_asset`."""
    patch_fero_get.return_value = asset_data
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    asset = client.get_asset("some-uuid")
    assert isinstance(asset, Asset)
    assert asset.name == asset_data["name"]
    patch_fero_get.assert_called_with("/api/assets/some-uuid/")


def test_search_assets(patch_fero_get, asset_data):
    """Test that an iterator of assets is returned by `search_assets`."""
    patch_fero_get.return_value = {"next": None, "results": [asset_data]}
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    assets = [a for a in client.search_assets(asset_data["name"])]
    assert len(assets) == 1
    assert isinstance(assets[0], Asset)
    assert assets[0].name == asset_data["name"]
    patch_fero_get.assert_called_with(
        "/api/assets/", params={"name": asset_data["name"]}
    )


def test_search_assets_paginated(patch_fero_get, asset_data):
    """Test that an iterator of assets is returned by `search_assets`."""
    patch_fero_get.side_effect = [
        {
            "next": "http://test.com/api/assets/?token=1234",
            "results": [asset_data],
        },
        {"next": None, "results": [asset_data]},
    ]
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    assets = [a for a in client.search_assets(asset_data["name"])]
    assert len(assets) == 2
    assert isinstance(assets[0], Asset)
    assert assets[0].name == asset_data["name"]
    assert isinstance(assets[1], Asset)
    assert assets[1].name == asset_data["name"]
    patch_fero_get.assert_has_calls(
        [
            mock.call("/api/assets/", params={"name": asset_data["name"]}),
            mock.call("/api/assets/?token=1234", params=None),
        ]
    )


def test_get_datasource_success(patch_fero_get, datasource_data):
    """Test that an data source is returned by `get_datasource`."""
    patch_fero_get.return_value = datasource_data
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    ds = client.get_datasource("9f79206e-94fc-4834-8f52-84008b12df86")
    assert isinstance(ds, DataSource)
    assert ds.name == datasource_data["name"]
    patch_fero_get.assert_called_with(
        "/api/v2/data_source/9f79206e-94fc-4834-8f52-84008b12df86/"
    )


def test_get_process_success(patch_fero_get, process_data):
    """Test that an process is returned by `get_process`."""
    patch_fero_get.return_value = process_data
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    process = client.get_process("some-uuid")
    assert isinstance(process, Process)
    assert process.name == process_data["name"]
    patch_fero_get.assert_called_with("/api/processes/some-uuid/")


def test_search_processes(patch_fero_get, process_data):
    """Test that the correct iterator of processes is returned by `search_processes`."""
    patch_fero_get.return_value = {"next": None, "results": [process_data]}
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    processes = [a for a in client.search_processes(process_data["name"])]
    assert len(processes) == 1
    assert isinstance(processes[0], Process)
    assert processes[0].name == process_data["name"]
    patch_fero_get.assert_called_with(
        "/api/processes/", params={"name": process_data["name"]}
    )


def test_search_processes_paginated(patch_fero_get, process_data):
    """Test that a list process is returned by `search_analyses`."""
    patch_fero_get.side_effect = [
        {
            "next": "http://test.com/api/processes/?token=1234",
            "results": [process_data],
        },
        {"next": None, "results": [process_data]},
    ]
    client = Fero(fero_token="fakeToken", hostname="http://test.com")
    processes = [a for a in client.search_processes(process_data["name"])]
    assert len(processes) == 2
    assert isinstance(processes[0], Process)
    assert processes[0].name == process_data["name"]
    assert isinstance(processes[1], Process)
    assert processes[1].name == process_data["name"]
    patch_fero_get.assert_has_calls(
        [
            mock.call("/api/processes/", params={"name": process_data["name"]}),
            mock.call("/api/processes/?token=1234", params=None),
        ]
    )



"""A module to test `DataSource` and related classes."""

import pytest
from fero import FeroError
from fero.datasource import UploadedFileStatus


@pytest.fixture
def file_status_data():
    """Get sample data matching the `UploadedFileStatus` class structure."""
    return {
        "uuid": "5351ab61-a50b-428d-adbb-81c927a8f14c",
        "status": "D",
        "error_notices": {"global_notices": [], "parsing_notices": []},
    }


def test_get_upload_status_makes_correct_calls(patched_fero_client, file_status_data):
    """Test that `get_upload_status` makes the correct call to the server."""
    file_uuid = file_status_data["uuid"]
    file_status = UploadedFileStatus(patched_fero_client, file_uuid)
    patched_fero_client.get.return_value = file_status_data

    status = file_status.get_upload_status()
    status["uuid"] = str(status["uuid"])
    assert status == file_status_data
    patched_fero_client.get.assert_called_with(
        f"/api/v2/uploaded_files/{file_uuid}/", allow_404=True
    )


def test_wait_until_complete_returns_good(patched_fero_client, file_status_data):
    """Test that `wait_until_complete` returns the instance when complete."""
    file_uuid = file_status_data["uuid"]
    file_status = UploadedFileStatus(patched_fero_client, file_uuid)
    patched_fero_client.get.return_value = file_status_data

    assert isinstance(file_status.wait_until_complete(), UploadedFileStatus)


def test_wait_until_complete_raises_error(patched_fero_client, file_status_data):
    """Test that `wait_until_complete` raises the returned error."""
    file_uuid = file_status_data["uuid"]
    file_status = UploadedFileStatus(patched_fero_client, file_uuid)
    file_status_data["error_notices"]["parsing_notices"] = [{"kind": "test_error"}]
    file_status_data["status"] = "E"
    patched_fero_client.get.return_value = file_status_data
    with pytest.raises(FeroError) as e:
        file_status.wait_until_complete()

    assert (
        e.value.message
        == f'Unable to upload file.  The following error(s) occurred: "{str({"kind": "test_error"})}"'
    )



"""This Module holds the Fero client class, which is used to communicate with the Fero API."""

from fero.datasource import DataSource
import os
import io
import re
import requests
from urllib.parse import urlparse
from azure.storage.blob import BlobClient
from pathlib import Path
from typing import Dict, Iterator, Optional, Union, Type
from . import FeroError
from .analysis import Analysis
from .asset import Asset
from .process import Process
from .common import FeroObject

FERO_CONF_FILE = ".fero"


class Fero:
    """The basic client used for communicating the the Fero API."""

    def __init__(
        self,
        hostname: Optional[str] = "https://app.ferolabs.com",
        username: Optional[str] = None,
        password: Optional[str] = None,
        fero_token: Optional[str] = None,
        verify: bool = True,
    ):
        """Create a base client for communicating with the Fero API.

        This class uses a JWT to query the api in all cases, however it will attempt to obtain a token if a username
        and password are provided.  It attempts to procure the token in the following order.

        1.  A token is provided in the constructor.
        2.  A username and password are provided in the constructor.
        3.  A token is provided in the "FERO_TOKEN" env variable.
        4.  A token is defined in the user's .fero file.
        5.  A username and password are provided in the environment variables "FERO_PASSWORD" and "FERO_USERNAME".
        6.  A username and password are provided in the .fero file.

        :param hostname: URL of the Fero server, defaults to "https://app.ferolabs.com/"
        :type hostname: Optional[str], optional
        :param username: username to user to login, defaults to None
        :type username: Optional[str], optional
        :param password: password to use to login, defaults to None
        :type password: Optional[str], optional
        :param fero_token: JWT token to use, defaults to None
        :type fero_token: Optional[str], optional
        :param verify: whether requests should verify ssl, defaults to True
        :type verify: bool
        :raises FeroError: Raised if a token cannot be obtained
        """
        self._fero_token = None
        self._fero_conf_content = None
        self._password = None
        self._username = None
        self._verify = verify

        self._hostname = hostname.rstrip("/")

        if fero_token:
            self._fero_token = fero_token

        if username and password:
            self._username = username
            self._password = password
            self._fero_token = self._get_token_as_user()

        if self._fero_token is None:
            self._fero_token = self._get_token_string_from_local()

        if self._fero_token is None:
            self._fero_token = self._jwt_from_local_user_pass()

        if self._fero_token is None:
            raise FeroError("Could not login into Fero")

    @property
    def fero_conf_contents(self) -> str:
        """Cache the file content."""
        if self._fero_conf_content is None:
            self._get_fero_conf_contents()

        return self._fero_conf_content

    def _get_token_string_from_local(self) -> Union[str, None]:
        """Check the local system for the JWT token string."""
        if "FERO_TOKEN" in os.environ:
            return os.environ.get("FERO_TOKEN")
        jwt_match = re.search(r"FERO_TOKEN=([^\n]+)\n", self.fero_conf_contents)
        return jwt_match.group(1) if jwt_match else None

    def _get_fero_conf_contents(self) -> str:
        """Load the content of a .fero file in the user's home directory."""
        home = Path.home()
        fero_conf = home / FERO_CONF_FILE
        if fero_conf.is_file():
            with open(str(fero_conf)) as conf_file:
                self._fero_conf_content = conf_file.read()
        else:
            self._fero_conf_content = ""

    def _get_token_as_user(self) -> Union[str, None]:
        """Get the token with the username and password."""
        req = requests.post(
            f"{self._hostname}/api/token/auth/",
            json={"username": self._username, "password": self._password},
            verify=self._verify,
        )

        return req.json().get("token", None)

    def _jwt_from_local_user_pass(self) -> Union[str, None]:
        """Check the local system for username and password and queries fero for a JWT."""
        if "FERO_USERNAME" in os.environ and "FERO_PASSWORD" in os.environ:
            self._username = os.environ.get("FERO_USERNAME")
            self._password = os.environ.get("FERO_PASSWORD")

        if self._username is None or self._password is None:

            username_match = re.search(
                r"FERO_USERNAME=([^\n]+)\n", self.fero_conf_contents
            )
            password_match = re.search(
                r"FERO_PASSWORD=([^\n]+)\n", self.fero_conf_contents
            )

            if username_match and password_match:
                self._username = username_match.group(1)
                self._password = password_match.group(1)

        if self._username and self._password:
            return self._get_token_as_user()

        return None

    @staticmethod
    def _handle_response(
        response: requests.Response, allow_404: bool = False
    ) -> Optional[Union[dict, bytes]]:
        """Check and decode a request response and raise a relevant error if needed."""
        if 200 <= response.status_code < 300:
            if response.headers.get("content-type") == "application/json":
                return response.json()
            else:
                return response.content
        elif response.status_code == 404:
            if allow_404:
                return None
            else:
                raise FeroError("The requested resource was not found.")
        elif response.status_code in [401, 403]:
            raise FeroError("You are not authorized to access this resource.")
        else:
            raise FeroError(
                f"There was an issue connecting to Fero. - Status Code: {response.status_code}"
            )

    def _s3_upload(self, inbox_response, file_name, fp) -> None:

        files = {
            "file": (
                file_name,
                fp,
            )
        }
        res = requests.post(
            inbox_response["url"],
            data=inbox_response["fields"],
            files=files,
            verify=self._verify,
        )

        if res.status_code != 204:
            raise FeroError("Error Uploading File")

    @staticmethod
    def _azure_upload(inbox_response: dict, fp) -> None:
        blob_client = BlobClient.from_blob_url(
            f"https://{inbox_response['storage_name']}.blob.core.windows.net/{inbox_response['container']}/{inbox_response['blob']}?{inbox_response['sas_token']}"
        )
        blob_client.upload_blob(io.BytesIO(fp.read().encode()))

    def _paginated_get(
        self, url: str, object_class: Type[FeroObject], params: Dict[str, str]
    ) -> Iterator[FeroObject]:

        next_url = url

        while next_url is not None:
            parsed_url = urlparse(next_url)

            # we use the path here in case the server has custom ports behind a load balancer or
            # proxy
            next_url = parsed_url.path
            if params is None and parsed_url.query != "":
                next_url += f"?{parsed_url.query}"
            data = self.get(next_url, params=params)
            for result in data.get("results", []):
                yield object_class(self, result)

            next_url = data.get("next", None)
            # further params should come from the next_url
            params = None

    def upload_file(self, inbox_response, file_name, file_pointer):
        """Upload a file to object store specified by Fero inbox response."""
        if inbox_response["upload_type"] == "azure":
            self._azure_upload(inbox_response, file_pointer)
        else:
            self._s3_upload(inbox_response, file_name, file_pointer)

    def search_analyses(self, name: str = None) -> Iterator[Analysis]:
        """Search available analyses by name and return an iterator of matching objects.

        :param name: Name of analysis to filter by.
        :type name: str, optional
        :return: a list of analyses
        :rtype: Iterator[Analysis]
        """
        params = {}
        if name is not None:
            params["name"] = name
        return self._paginated_get("/api/analyses/", Analysis, params=params)

    def get_analysis(self, uuid: str) -> Analysis:
        """Get a Fero Analysis using the UUID.

        :param uuid: UUID of the analysis
        :type uuid: str
        :return: An Analysis object
        :rtype: Analysis
        """
        analysis_data = self.get(f"/api/analyses/{uuid}/")
        return Analysis(self, analysis_data)

    def search_assets(self, name: str = None) -> Iterator[Asset]:
        """Search available assets by name and return an iterator of matching objects.

        :param name: Name of asset to filter by.
        :type name: str, optional
        :return: a list of assets
        :rtype: Iterator[Asset]
        """
        params = {}
        if name is not None:
            params["name"] = name
        return self._paginated_get("/api/assets/", Asset, params=params)

    def get_asset(self, uuid: str) -> Asset:
        """Get a Fero Asset using the UUID.

        :param uuid: UUID of the asset
        :type uuid: str
        :return: An Asset object
        :rtype: Asset
        """
        asset_data = self.get(f"/api/assets/{uuid}/")
        return Asset(self, asset_data)

    def search_processes(self, name: str = None) -> Iterator[Process]:
        """Search available processes by name and return an iterator of matching objects.

        :param name: Name of analysis to filter by.
        :type name: str, optional
        :return: a list of processes
        :rtype: Iterator[Process]
        """
        params = {}
        if name is not None:
            params["name"] = name
        return self._paginated_get("/api/processes/", Process, params=params)

    def get_process(self, uuid: str) -> Process:
        """Get a Fero Process using the UUID.

        :param uuid: UUID of the process
        :type uuid: str
        :return: An Process object
        :rtype: Analysis
        """
        process_data = self.get(f"/api/processes/{uuid}/")
        return Process(self, process_data)

    def get_datasource(self, uuid: str) -> DataSource:
        """Get a Fero Data Source by UUID.

        :param uuid: UUID of requested object
        :type uuid: str
        :return: A data source object
        :rtype: DataSource
        """
        source_data = self.get(f"/api/v2/data_source/{uuid}/")
        return DataSource(self, source_data)

    def post(self, url: str, data: dict) -> Union[dict, bytes]:
        """Do a POST request with headers set."""
        return self._handle_response(
            requests.post(
                f"{self._hostname}{url}",
                json=data,
                headers={"Authorization": f"JWT {self._fero_token}"},
                verify=self._verify,
            ),
            allow_404=False,
        )

    def get(self, url: str, params=None, allow_404=False) -> Union[dict, bytes]:
        """Do a GET request with headers set."""
        return self._handle_response(
            requests.get(
                f"{self._hostname}{url}",
                params=params,
                headers={"Authorization": f"JWT {self._fero_token}"},
                verify=self._verify,
            ),
            allow_404=allow_404,
        )

    def get_preauthenticated(self, url, params=None) -> Union[dict, bytes]:
        """Do a GET request without adjusting the url or auth headers."""
        return self._handle_response(
            requests.get(
                url,
                params=params,
                verify=self._verify,
            )
        )



"""This Module defines classes related to a fero datasource."""

import os
import time

import requests
import fero
from fero import FeroError
from typing import Optional, Union
from marshmallow import (
    Schema,
    fields,
    validate,
    EXCLUDE,
)

from .common import FeroObject

# 1 Mb
CHUNK_SIZE = 1048576


class DataSourceSchema(Schema):
    """A schema to store data related to a fero datasource."""

    class Meta:
        """
        Specify that unknown fields included on this schema should be excluded.

        See https://marshmallow.readthedocs.io/en/stable/quickstart.html#handling-unknown-fields.
        """

        unknown = EXCLUDE

    uuid = fields.UUID(required=True)

    # deprecated in favor of primary_key_columns
    primary_key_column = fields.String(required=False, allow_none=True)
    primary_key_columns = fields.List(fields.String, required=False, allow_none=True)
    primary_datetime_column = fields.String(required=True, allow_none=True)

    schema = fields.Dict(required=True, allow_none=True)

    name = fields.String(required=True)
    description = fields.String(required=True)

    created = fields.DateTime(required=True)
    modified = fields.DateTime(required=True)
    ac_name = fields.String(required=True)
    username = fields.String(required=True)

    INITIALIZED = "I"
    PROCESSING = "P"
    LOADING_FILE = "L"
    ANALYZING_FILE = "A"
    WRITING_FILE = "W"
    COMPRESSING_TABLE = "C"
    READY = "R"
    ERROR = "E"
    status = fields.String(
        validate=validate.OneOf(["I", "P", "L", "A", "W", "C", "R", "E"]), required=True
    )

    error_notices = fields.Dict(required=True, default=lambda: {"errors": []})
    progress = fields.Integer(required=True, default=0)

    overwrites = fields.Dict(required=True, allow_none=True)

    transformed_source = fields.Bool(required=True, default=False)
    live_source = fields.Bool(required=True, default=False)
    default_upload_config = fields.Dict(required=False)
    raw_file = fields.Url(required=False, allow_none=True)
    download_url = fields.Url(required=False, allow_none=True)


class DataSource(FeroObject):
    """An object for interacting with a specific Data Source on Fero.

    Adding data to a Data Source allows Fero to interface with that data.
    This data can then be added to a Fero Process and later analyzed in
    an Asset or Analysis.
    """

    schema_class = DataSourceSchema

    def __repr__(self) -> str:
        """Represent the `DataSource` by its name."""
        return f"<Data Source name={self.name}>"

    __str__ = __repr__

    def append_csv(self, file_path: str, wait_until_complete: bool = False):
        """Append a specified csv file to the data source.

        :param file_path: Location of the csv file to append
        :type file_path: str
        :raises FeroError: Raised if the file does not match a naive csv check
        """
        if not file_path.endswith(".csv"):
            raise FeroError("Fero only supports csv appends")

        file_name = os.path.basename(file_path)

        inbox_response = self._client.post(
            f"/api/v2/data_source/{self.uuid}/inbox_url/",
            {"file_name": file_name, "action": "A"},
        )
        with open(file_path) as fp:
            self._client.upload_file(inbox_response, file_name, fp)

        upload_status = UploadedFileStatus(self._client, inbox_response["upload_uuid"])
        return (
            upload_status.wait_until_complete()
            if wait_until_complete
            else upload_status
        )

    def replace_csv(self, file_path: str, wait_until_complete: bool = False):
        """Append a specified csv file to the data source.

        :param file_path: Location of the csv file to append
        :type file_path: str
        :raises FeroError: Raised if the file does not match a naive csv check
        """
        if not file_path.endswith(".csv"):
            raise FeroError("Fero only supports csv appends")

        file_name = os.path.basename(file_path)

        inbox_response = self._client.post(
            f"/api/v2/data_source/{self.uuid}/inbox_url/",
            {"file_name": file_name, "action": "R"},
        )
        with open(file_path) as fp:
            self._client.upload_file(inbox_response, file_name, fp)

        upload_status = UploadedFileStatus(self._client, inbox_response["upload_uuid"])

        return (
            upload_status.wait_until_complete()
            if wait_until_complete
            else upload_status
        )

    def get_download_status(self) -> Optional[str]:
        """Get current status of the data source download.

        :return: String if file download is completely processed, None if still processing
        """
        data = self._client.get(f"/api/v2/data_source/{self.uuid}/")
        self._data = self.schema_class().load(data)
        return self.download_url

    def wait_until_download_complete(self):
        """Continuously check the status of the download until it succeeds or fails."""
        status = self.get_download_status()
        while not status:
            time.sleep(0.5)
            status = self.get_download_status()

    def download(self, raw=False) -> str:
        """Download data source data as a csv.

        :param raw: Whether to download the raw data
        :raises FeroError: Raised if the raw download is not available
        :return: The local filename of the download
        """
        if raw:
            url = self.raw_file
            if url is None:
                raise FeroError("Raw download for that datasource is not available.")

        else:
            url = self.download_url
            if url is None:
                self._client.get(f"/api/v2/data_source/{self.uuid}/create_download/")
                self.wait_until_download_complete()
                url = self.download_url

        # Download the file and write as a stream since it could be large
        req = requests.get(url, stream=True)
        download_filename = f'fero{"-raw" if raw else ""}-ds-{self.uuid}.csv'
        with open(download_filename, "wb") as fp:
            for chunk in req.iter_content(chunk_size=CHUNK_SIZE):
                fp.write(chunk)

        return download_filename


class UploadedFilesSchema(Schema):
    """A schema to store data related to files uploaded to fero."""

    class Meta:
        """
        Specify that unknown fields included on this schema should be excluded.

        See https://marshmallow.readthedocs.io/en/stable/quickstart.html#handling-unknown-fields.
        """

        unknown = EXCLUDE

    uuid = fields.UUID(required=True)

    INITIALIZED = "I"
    PARSING = "P"
    ANALYZING = "A"
    CREATING = "C"
    ERROR = "E"
    DONE = "D"
    DELETING = "R"  # R for removing
    USER_CONFIRMATION = "U"

    status = fields.String(
        validate=validate.OneOf(["I", "P", "A", "C", "R", "D", "E", "U"]), required=True
    )

    error_notices = fields.Dict(
        required=True, default=lambda: {"global_notices": [], "parsing_notices": []}
    )


class UploadedFileStatus:
    """An object to track the status of a file during its upload."""

    def __init__(self, client: "fero.Fero", id: str):
        """Create an `UploadedFileStatus` from the uplolad ID to track the file's upload status."""
        self._id = id
        self._client = client
        self._status_data = None
        self._schema = UploadedFilesSchema()

    @staticmethod
    def _check_status_complete(status: Optional[dict]) -> bool:
        """Check status of the latest uploaded file response.

        Returns true if complete, false if not complete and raises an error if the status is error.
        """
        if status is None or status["status"] not in [
            UploadedFilesSchema.ERROR,
            UploadedFilesSchema.DONE,
        ]:
            return False

        if status["status"] == UploadedFilesSchema.ERROR:

            errors = [
                f'"{str(e)}"'
                for e in status["error_notices"]["global_notices"]
                + status["error_notices"]["parsing_notices"]
            ]

            error_message = f"Unable to upload file.  The following error(s) occurred: {', '.join(errors)}"
            raise FeroError(error_message)

        return True

    def get_upload_status(self) -> Union[dict, None]:
        """Get current status of the uploaded files object.

        :return: Dict if it is completely processed, None if still processing
        :raises FeroError: Raised if fero was unable to process the file
        """
        raw_data = self._client.get(
            f"/api/v2/uploaded_files/{self._id}/", allow_404=True
        )
        data = None
        if raw_data is not None:
            data = self._schema.load(raw_data)

        return data

    def wait_until_complete(self) -> "UploadedFileStatus":
        """Continuously check the status of the upload until it succeeds or fails."""
        status = self.get_upload_status()
        while not self._check_status_complete(status):
            time.sleep(0.5)
            status = self.get_upload_status()

        return self



"""This module defines classes related to a fero process."""

import time
import fero
import requests
from tempfile import NamedTemporaryFile
from functools import lru_cache as memoized
from typing import Sequence, Optional, Union
from .common import FeroObject, poll
from marshmallow import (
    Schema,
    fields,
    validate,
    EXCLUDE,
)
import pandas as pd
from .exceptions import FeroError

# 1 Mb
CHUNK_SIZE = 1048576


class DataRequestError(FeroError):
    """Raised when data cannot be generated."""

    pass


class FeroProcessTypes:
    """An Enum class representing the type of fero process."""

    ADVANCED = "A"
    BATCH = "B"
    CONTINUOUS = "C"

    @classmethod
    def validator(cls):
        """Validate that a given process type is one of the possible options."""
        return validate.OneOf([cls.ADVANCED, cls.BATCH, cls.CONTINUOUS])


class SnapShotStatus:
    """An Enum class representing the status of a process snpashot."""

    READY = "R"
    INITIALIZED = "I"
    PROCESSING = "P"
    ERROR = "E"

    @classmethod
    def validator(cls):
        """Validate that a given snapshot status is one of the possible options."""
        return validate.OneOf([cls.READY, cls.INITIALIZED, cls.PROCESSING, cls.ERROR])


class TagGroupCategory:
    """An Enum class representating the strength of a tag group."""

    STRONG = "s"
    BORDERLINE = "b"
    WEAK = "w"

    @classmethod
    def validator(cls):
        """Validate that a given tag group strength is one of the possible options."""
        return validate.OneOf([cls.STRONG, cls.BORDERLINE, cls.WEAK])


class TagSchema(Schema):
    """A schema to store data related to a fero process tag."""

    class Meta:
        """
        Specify that unknown fields included on this schema should be excluded.

        See https://marshmallow.readthedocs.io/en/stable/quickstart.html#handling-unknown-fields.
        """

        unknown = EXCLUDE

    name = fields.String(required=True)
    description = fields.String(required=True)
    display_format = fields.Integer(required=True, allow_none=True)
    cost = fields.Float(required=True, allow_none=True)
    unit = fields.String(required=True)
    alias = fields.String(required=True)
    minimum = fields.Dict(required=True, allow_none=True)
    maximum = fields.Dict(required=True, allow_none=True)
    lower_limit = fields.Dict(required=True, allow_none=True)
    upper_limit = fields.Dict(required=True, allow_none=True)
    step_size = fields.Float(required=True, allow_none=True)
    disabled = fields.Boolean(required=True)
    proxy = fields.Integer(required=True)


class TagGroupSchema(Schema):
    """A schema to store data related to fero stage tag groups."""

    class Meta:
        """
        Specify that unknown fields included on this schema should be excluded.

        See https://marshmallow.readthedocs.io/en/stable/quickstart.html#handling-unknown-fields.
        """

        unknown = EXCLUDE

    id = fields.Integer(required=True)
    category = fields.String(required=True, validate=TagGroupCategory.validator())
    factor_scores = fields.Dict(required=True)
    representative_tag = fields.Integer(required=True)
    tags = fields.List(fields.Integer, required=True)


class StageSchema(Schema):
    """A schema to store data related to a fero process stage."""

    class Meta:
        """
        Specify that unknown fields included on this schema should be excluded.

        See https://marshmallow.readthedocs.io/en/stable/quickstart.html#handling-unknown-fields.
        """

        unknown = EXCLUDE

    id = fields.Integer(required=True)
    proxy = fields.Integer(required=True)
    name = fields.String(required=True)
    order = fields.Integer(required=True)
    tags = fields.List(fields.Nested(TagSchema), required=True)
    configuration = fields.Dict(required=True)


class NestedSnapshotSchema(Schema):
    """A schema to store data related to a fero process snapshot."""

    uuid = fields.String(required=True)
    status = fields.String(
        required=True,
        validate=SnapShotStatus.validator(),
    )


class ProcessSchema(Schema):
    """A schema to store data related to a fero process."""

    class Meta:
        """
        Specify that unknown fields included on this schema should be excluded.

        See https://marshmallow.readthedocs.io/en/stable/quickstart.html#handling-unknown-fields.
        """

        unknown = EXCLUDE

    api_id = fields.String(required=True)
    name = fields.String(required=True)
    created = fields.DateTime(require=True)
    modified = fields.DateTime(require=True)
    latest_revision_version = fields.Integer(required=True)
    username = fields.String(required=True)
    process_type = fields.String(required=True, validate=FeroProcessTypes.validator())
    product_type = fields.String(Required=False, allow_none=True)
    kind = fields.String(required=True, validate=validate.OneOf(["process"]))
    latest_ready_snapshot = fields.Nested(NestedSnapshotSchema, allow_none=True)
    # TODO update this to a nested schema once the configuration settles more
    data_config = fields.Dict(required=True)


class Tag(FeroObject):
    """High level object representing a process tag.

    A tag is a single measurement type for a process and specifies metadata associated with this measurement such
    as limits and cost.
    """

    schema_class = TagSchema

    def __init__(self, process: "Process", client: "fero.Fero", data: dict):
        """
        Create a new representation of a Tag from the given data.

        :param process: The process this tag belongs to
        :param client: The fero client object that gives access to the `process`
        :param data: A dictionary holding values for the fields of a `TagSchema`
        """
        self._process = process
        super().__init__(client, data)

    def __repr__(self) -> str:
        """Represent the `Tag` by its name and its `Process`."""
        return f"<Tag name={self.name} Process={self._process}>"

    __str__ = __repr__

    def __eq__(self, other: object) -> bool:
        """Check whether `other` is a Tag with the same data as this tag."""
        return isinstance(other, Tag) and self._data == other._data

    @classmethod
    def tag_name(cls, instance: Union["Tag", str]) -> str:
        """
        Get the name of a given tag.

        :param instance: Either the `Tag` or the string name of the tag
        :return: The name of the specified tag
        """
        return instance if isinstance(instance, str) else instance.name


class TagGroup(FeroObject):
    """High level object representing a stage tag group."""

    schema_class = TagGroupSchema

    def __init__(self, stage: "Stage", client: "fero.Fero", data: dict):
        """
        Create a new representation of a Stage tag group from the given data.

        :param stage: The stage this tag group belongs to
        :param client: The fero client object that gives access to the `stage`
        :param data: A dictionary holding values for the fields of a `TagGroupSchema`
        """
        self._stage = stage
        super().__init__(client, data)
        self._data["tags"] = [
            next(tag for tag in self._stage.tags if tag.proxy == tag_proxy)
            for tag_proxy in self._data["tags"]
        ]
        self._data["representative_tag"] = next(
            tag
            for tag in self._stage.tags
            if tag.proxy == self._data["representative_tag"]
        )

    def __repr__(self) -> str:
        """Represent the `Tag Group` by its id and `Stage`."""
        return f"<TagGroup id={self.id} Stage={self._stage}>"

    __str__ = __repr__


class Stage(FeroObject):
    """High level object representing a process stage.

    A tag is a discrete grouping of related tags that are part of process and describes metadata about how to process each tag
    based on the type of process and configuration of the stage.
    """

    schema_class = StageSchema

    def __init__(self, process: "Process", client: "fero.Fero", data: dict):
        """
        Create a new representation of a Stage from the given data.

        :param process: The process this stage belongs to
        :param client: The fero client object that gives access to the `process`
        :param data: A dictionary holding values for the fields of a `StageSchema`
        """
        self._process = process
        super().__init__(client, data)
        self._data["tags"] = [
            Tag(process, client, tdata) for tdata in self._data["tags"]
        ]

    def __repr__(self) -> str:
        """Represent the `Stage` by its name and its `Process`."""
        return f"<Stage name={self.name} Process={self._process}>"

    __str__ = __repr__

    def __eq__(self, other) -> bool:
        """Check whether `other` is a Stage with the same data as this stage."""
        return isinstance(other, Stage) and self._data == other._data

    @property
    def tag_names(self) -> Sequence[str]:
        """Get the names of the tags in the stage.

        :return: A list of tag names in the stage
        """
        return [Tag.tag_name(t) for t in self.tags]

    @property
    def tag_groups(self) -> Sequence["TagGroup"]:
        """Get the tag groups for this process and stage from Fero.

        Request is memoized to avoid unnecessary additional calls.

        :return: A List of tag groups
        """
        url = (
            f"/api/processes/{self._process.api_id}/tag_groups/?stageProxy={self.proxy}"
        )
        res = self._client.get(url)
        if res.get("computing") and not res.get("error"):
            time.sleep(0.5)
            res = self._client.get(url)

        return [
            TagGroup(self, self._client, tg_data) for tg_data in res.get("tag_groups")
        ]


class Process(FeroObject):
    """High level object for interacting with a Process object defined in Fero.

    A Process is a set of configurations that describe an industrial process and allow
    Fero to combine, transform, and enrich raw data sources before the final data is
    used in an Analysis.
    """

    schema_class = ProcessSchema

    def __repr__(self) -> str:
        """Represent the `Process` by its name."""
        return f"<Process name={self.name}>"

    __str__ = __repr__

    @property
    @memoized(maxsize=1)
    def stages(self) -> Sequence[Stage]:
        """Get the stages for this process from Fero.

        Request is memoized to avoid unnecessary additional calls.

        :return: A List of stages
        """
        return [
            Stage(self, self._client, stage_data)
            for stage_data in self._client.get(
                f"/api/processes/{self.api_id}/stages/"
            ).get("stages")
        ]

    @property
    @memoized(maxsize=1)
    def tags(self) -> Sequence[Tag]:
        """Get the tags for this process from Fero.

        Request is memoized to avoid unnecessary additional calls.

        :return: List of tags associated with the process
        """
        return [
            Tag(self, self._client, tag_data)
            for tag_data in self._client.get(f"/api/processes/{self.api_id}/tags/").get(
                "tags"
            )
        ]

    def _tag_stage_index(self, tag: Union[Tag, str]) -> Union[int, None]:

        stage_idx = None
        tag_name = Tag.tag_name(tag)
        for idx, s in enumerate(self.stages):
            if tag_name in s.tag_names:
                stage_idx = idx
                break

        return stage_idx

    def get_tag_stage(self, tag: Union[Tag, str]) -> Union[Stage, None]:
        """Return the process stage specified in the stage or None if the tag is not found.

        :param tag: Tag to search for
        :return: [description]
        """
        stage_idx = self._tag_stage_index(tag)
        return None if stage_idx is None else self.stages[stage_idx]

    def get_stages_by_kpis(self, kpis: Sequence[Union[Tag, str]]) -> Sequence[Stage]:
        """Get a list of stages that would be used by a process for the given kpis.

        :param kpis: List of kpi tags
        :return: a list of stages
        """
        idxs = [self._tag_stage_index(k) for k in kpis]
        idxs = [i for i in idxs if i is not None]

        return [] if len(idxs) == 0 else self.stages[: max(idxs) + 1]

    def get_data(
        self,
        tags: Sequence[Union[Tag, str]],
        kpis: Optional[Sequence[Union[Tag, str]]] = None,
    ) -> pd.DataFrame:
        """Download a pandas data frame with specified process data.

        Returns a pandas data frame consisting of the specified tags and bounded by an optional list of kpis.
        For all process types specifying a kpi will remove any tags from stages that are after
        the stage the kpi is assigned to.  For advanced and batch process types not providing a kpi will
        get tags from all stages.  For a continuous process, a kpi must be specified.

        :param tags: List of Tags or tag names to include in the data frame
        :type tags: Sequence[Union[Tag, str]]
        :param kpis: List of tags or tag names to use for kpis, defaults to None
        :type kpis: Optional[Sequence[Union[Tag, str]]], optional
        :raises DataRequestError: Raised if data can not be generated
        :return: A data frame of the data generated by the process.
        :rtype: pd.DataFrame
        """
        if self.process_type == FeroProcessTypes.CONTINUOUS and kpis is None:
            raise DataRequestError(
                "A kpi must be specified to download continuous process data"
            )

        tag_names = [Tag.tag_name(tag) for tag in tags]
        kpi_names = None
        if kpis:
            kpi_names = [tag if isinstance(tag, str) else tag.name for tag in kpis]

        initial_response = self._client.post(
            f"/api/processes/{self.api_id}/download_process_data/",
            {"request_data": {"tags": tag_names, "kpis": kpi_names}},
        )

        request_id = initial_response["request_id"]

        def get_data():
            return self._client.post(
                f"/api/processes/{self.api_id}/download_process_data/",
                {"request_id": request_id},
            )

        download_response = poll(get_data, lambda d: d["complete"])

        download_data = download_response["data"]

        if not download_data["success"]:
            raise DataRequestError(download_data["message"])

        # Download the file and write as a stream since it could be large
        req = requests.get(download_data["download_url"], stream=True)
        req.raise_for_status()

        with NamedTemporaryFile() as fp:
            for chunk in req.iter_content(chunk_size=CHUNK_SIZE):
                fp.write(chunk)
            fp.seek(0)
            df = pd.read_parquet(fp.name)
        return df



"""A module to test `Analysis` and related classes."""

import io
import datetime
from uuid import uuid4
import pytest
from fero import FeroError
from unittest import mock
from fero.analysis import Analysis, Prediction
import pandas as pd
from pandas.testing import assert_frame_equal


@pytest.fixture
def test_analysis(
    analysis_data,
    v1_interpreted_schema,
    patched_fero_client,
    prediction_request_response,
    prediction_results_response_completed,
):
    """Create a sample `Analysis` object."""
    patched_fero_client.post.return_value = prediction_request_response
    patched_fero_client.get.return_value = prediction_results_response_completed

    analysis = Analysis(patched_fero_client, analysis_data)
    analysis._schema_cache = v1_interpreted_schema
    return analysis


@pytest.fixture
def test_analysis_with_data(test_analysis):
    """Create a sample `Analysis` holding data for the prediction simulator."""
    test_analysis._presentation_data_cache = {
        "data": [
            {
                "id": "regression_simulator",
                "type": "regression_simulator",
                "title": "Prediction Simulator",
                "tab": "Prediction Simulator",
                "content": {
                    "factors": [
                        {
                            "factor": "Factor 1",
                            "min": 51.0569817398695,
                            "median": 148.02718539320762,
                            "max": 245.75370014648703,
                            "importances": [0.9184262466993418],
                            "step": 1.9469671840661755,
                            "default": 210.17636699824746,
                            "dtype": "float",
                        },
                        {
                            "factor": "Factor 2",
                            "min": 51.0569817398695,
                            "median": 148.02718539320762,
                            "max": 245.75370014648703,
                            "importances": [0.9184262466993418],
                            "step": 1.9469671840661755,
                            "default": 210.17636699824746,
                            "dtype": "float",
                        },
                        {
                            "factor": "Factor 3",
                            "min": 51.0569817398695,
                            "median": 148.02718539320762,
                            "max": 245.75370014648703,
                            "importances": [0.9184262466993418],
                            "step": 1.9469671840661755,
                            "default": 210.17636699824746,
                            "dtype": "float",
                        },
                        {
                            "factor": "Category 0",
                            "median": "4",
                            "importances": [0.0004481622416081965],
                            "default": "0",
                            "dtype": "category",
                            "categories": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
                        },
                    ],
                    "targets": [
                        {
                            "name": "Target 1",
                            "default_measured": 774.2382246688347,
                            "default": {
                                "low": [734.5873947705951],
                                "mid": [767.84871403268],
                                "high": [801.110033294765],
                            },
                            "min": 552.3692540713583,
                            "max": 819.9899001874504,
                        },
                        {
                            "name": "Target 2",
                            "default_measured": 774.2382246688347,
                            "default": {
                                "low": [734.5873947705951],
                                "mid": [767.84871403268],
                                "high": [801.110033294765],
                            },
                            "min": 552.3692540713583,
                            "max": 819.9899001874504,
                        },
                    ],
                },
            }
        ]
    }

    return test_analysis


@pytest.fixture
def expected_optimization_config():
    """Create a sample optimization configuration."""
    return {
        "name": "test optimization",
        "description": "",
        "prediction_type": "O",
        "input_data": {
            "kind": "OptimizationRequest",
            "version": 1,
            "useAdaptiveGrid": False,
            "objectives": [{"factor": "Factor 2", "goal": "maximize"}],
            "basisSpecifiedColumns": [],
            "linearFunctionDefinitions": {},
            "basisValues": {
                "Factor 1": 148.02718539320762,
                "Factor 2": 148.02718539320762,
                "Factor 3": 148.02718539320762,
                "Target 1": 767.84871403268,
                "Target 2": 767.84871403268,
                "Category 0": "4",
            },
            "bounds": [
                {
                    "factor": "Factor 1",
                    "lowerBound": 60.0,
                    "upperBound": 200.0,
                    "dtype": "factor_float",
                },
                {
                    "factor": "Target 1",
                    "lowerBound": 600.0,
                    "upperBound": 700.0,
                    "dtype": "target_float",
                    "confidenceInterval": "exclude",
                },
                {
                    "factor": "Factor 2",
                    "lowerBound": 70.0,
                    "upperBound": 100.0,
                    "dtype": "factor_float",
                },
            ],
        },
    }


@pytest.fixture
def prediction_request_response(expected_optimization_config):
    """Get a sample prediction request resopnse using the sample optimization configuration."""
    response = {
        "latest_prediction": {
            "uuid": "01bff6f2-8fb3-469e-813a-9b6cfd93e338",
            "created_by": {"id": 2, "username": "fero"},
            "created": "2020-09-30T12:35:44.897461Z",
            "modified": "2020-09-30T12:35:46.111257Z",
            "name": "g",
            "description": "",
            "prediction_type": "O",
            "progress_url": "/api/prediction_results/f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7/progress/",
            "latest_results": "f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7",
            "result_data": {},
            "ready": False,
            "prediction_tag": None,
            "url": "/api/analyses/21621466-b198-45be-89f9-3a5eb2c7cf48/predictions/01bff6f2-8fb3-469e-813a-9b6cfd93e338/",
            "latest_results_url": "/api/prediction_results/f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7/",
        }
    }

    response["input_data"] = expected_optimization_config["input_data"]

    return response


@pytest.fixture
def prediction_results_response_started():
    """Get sample results for a prediction when the prediction has started but not completed."""
    return {
        "request": "c9448486-0c59-487f-9c9c-345d98103fcb",
        "revision_model": "e751f70e-aeb4-46ee-b860-bfc37f3767a7",
        "result_data": {},
        "state": "P",
        "prediction_type": "O",
    }


@pytest.fixture
def prediction_results_response_completed(prediction_results_response_started):
    """Get sample results for a prediction when the prediction has completed."""
    prediction_results_response_started["result_data"] = {
        "status": "SUCCESS",
        "version": 1,
        "data": {
            "values": {
                "index": [0],
                "columns": [
                    "Factor 1",
                    "Factor 2",
                    "Factor 3",
                    "Target 1 (5%)",
                    "Target 1 (Mean)",
                    "Target 1 (95%)",
                    "Target 2 (5%)",
                    "Target 2 (Mean)",
                    "Target 2 (95%)",
                ],
                "data": [
                    [
                        148.02718539320762,
                        90.234,
                        148.02718539320762,
                        666.66666,
                        677.1234,
                        699.1234,
                        700.4567,
                        677.1234,
                        699.1234,
                    ],
                ],
            }
        },
        "kind": "OptimizationResponse",
    }

    prediction_results_response_started["state"] = "C"

    return prediction_results_response_started


@pytest.fixture
def prediction_data():
    """Get sample data returned by a prediction."""
    return [
        {"value1": 1.0, "value2": 2, "value3": 3.3},
        {"value1": 4.0, "value2": 5, "value3": 6.3},
    ]


@pytest.fixture
def prediction_dataframe(prediction_data):
    """Get sample data returned by a prediction in the form of a Dataframe."""
    return pd.DataFrame(prediction_data)


@pytest.fixture
def prediction_result_data_single_v1():
    """Get sample data returned by a prediction in the form of a Dataframe."""
    return {
        "status": "SUCCESS",
        "version": 1,
        "data": {
            "Target 1": {"value": {"high": [5.0], "low": [1.0], "mid": [3.0]}},
            "target2": {"value": {"high": [1.0], "low": [1.0], "mid": [1.0]}},
        },
    }


@pytest.fixture
def prediction_result_data_bulk_v1():
    """Get sample data returned by a v1 bulk prediction."""
    return {
        "status": "SUCCESS",
        "version": 1,
        "data": {
            "Target 1": {
                "value": {"high": [5.0, 5.1], "low": [1.0, 1.1], "mid": [3.0, 3.1]}
            },
            "target2": {
                "value": {"high": [1.0, 1.1], "low": [1.0, 1.1], "mid": [1.0, 1.1]}
            },
        },
    }


@pytest.fixture
def prediction_result_data_single_v2():
    """Get sample data returned by a v2 single prediction."""
    return {
        "status": "SUCCESS",
        "version": 2,
        "data": {
            "Target 1": {
                "value": {
                    "high90": [5.0],
                    "high50": [4.0],
                    "low90": [1.0],
                    "low50": [2.0],
                    "mid": [3.0],
                }
            },
            "target2": {
                "value": {
                    "high90": [1.0],
                    "high50": [1.0],
                    "low90": [1.0],
                    "low50": [1.0],
                    "mid": [1.0],
                }
            },
        },
    }


@pytest.fixture
def prediction_result_data_bulk_v2():
    """Get sample data returned by a v2 bulk prediction."""
    return {
        "status": "SUCCESS",
        "version": 2,
        "data": {
            "Target 1": {
                "value": {
                    "high90": [5.0, 5.1],
                    "high50": [4.0, 4.1],
                    "low90": [1.0, 1.1],
                    "low50": [2.0, 2.1],
                    "mid": [3.0, 3.1],
                }
            },
            "target2": {
                "value": {
                    "high90": [1.0, 1.1],
                    "high50": [1.0, 1.1],
                    "low90": [1.0, 1.1],
                    "low50": [1.0, 1.1],
                    "mid": [1.0, 1.1],
                }
            },
        },
    }


@pytest.fixture
def batch_prediction_success_data_cache():
    """Get sample data by a successful batch prediction."""
    return {
        "request": "request_id",
        "revision_model": "revision_model_id",
        "result_data": {
            "data": {
                "filePath": "file_path.json",
                "download_url": "http://download/url/with/expiring/access/token/file.json",
            },
            "version": 1,
            "status": "SUCCESS",
            "kind": "BatchPredictionResponse",
        },
        "state": "C",
        "prediction_type": "M",
    }


@pytest.fixture
def batch_prediction_failure_data_cache():
    """Get sample data by an unsuccessful batch prediction."""
    return {
        "request": "request_id",
        "revision_model": "revision_model_id",
        "result_data": {
            "status": "FAILED",
            "message": "Something broke!",
        },
        "state": "C",
        "prediction_type": "M",
    }


def test_creates_analysis_correctly(analysis_data, patched_fero_client):
    """Test that a valid analysis is created from valid data and a valid client."""
    analysis = Analysis(patched_fero_client, analysis_data)

    assert isinstance(analysis, Analysis)
    assert str(analysis) == "<Analysis name=quality-example-data>"


def test_has_trained_model_true(analysis_data, patched_fero_client):
    """Test that has_trained_model is true if there is a latest revision model."""
    analysis = Analysis(patched_fero_client, analysis_data)
    assert analysis.has_trained_model() is True


def test_has_trained_model_false(analysis_data, patched_fero_client):
    """Test that has_trained_model is true if there is no revision model."""
    analysis_data["latest_completed_model"] = None
    analysis = Analysis(patched_fero_client, analysis_data)
    assert analysis.has_trained_model() is False


def test_gets_v1_schema_for_v1_correctly(
    analysis_data, v1_interpreted_schema, patched_fero_client
):
    """Test that we fetch the schema with the datasource endpoint for V1 analyses."""
    analysis_data["latest_completed_model_schema"] = {}
    ds_uuid = analysis_data["data_source"]
    analysis = Analysis(patched_fero_client, analysis_data)
    patched_fero_client.get.return_value = v1_interpreted_schema
    assert analysis._schema == v1_interpreted_schema
    patched_fero_client.get.assert_called_with(
        f"/api/data_sources/{ds_uuid}/latest_interpreted_schema/"
    )


def test_gets_v1_schema_for_v2_correctly(
    analysis_data, v1_interpreted_schema, patched_fero_client
):
    """Test that we fetch the schema with the process endpoint for V2 analyses."""
    analysis_data["latest_completed_model_schema"] = {}
    dummy_uuid = uuid4()
    analysis_data["process"] = str(dummy_uuid)
    analysis_data["data_source"] = None
    analysis = Analysis(patched_fero_client, analysis_data)
    patched_fero_client.get.return_value = v1_interpreted_schema
    assert analysis._schema == v1_interpreted_schema
    patched_fero_client.get.assert_called_with(
        f"/api/processes/{str(dummy_uuid)}/v1_interpreted_schema/"
    )


def test_make_prediction_dictionaries(
    analysis_data,
    patched_fero_client,
    prediction_data,
    prediction_dataframe,
    batch_prediction_success_data_cache,
):
    """Test that make prediction returns expected response as a dictionary of predictions."""
    analysis = Analysis(patched_fero_client, analysis_data)
    analysis._upload_file = mock.MagicMock()
    analysis._upload_file.return_value = "test_workspace_id"
    analysis._poll_workspace_for_prediction = mock.MagicMock()

    # Mock prediction
    test_prediction = Prediction(patched_fero_client, "result_id")
    test_prediction._data_cache = batch_prediction_success_data_cache
    test_prediction._complete = True
    test_prediction.get_results = mock.MagicMock()
    test_prediction.get_results.return_value = prediction_dataframe  # As pd.DataFrame
    analysis._poll_workspace_for_prediction.return_value = test_prediction

    results = analysis.make_prediction(prediction_data)

    assert isinstance(results, list)
    assert results == [
        {"value1": 1.0, "value2": 2, "value3": 3.3},
        {"value1": 4.0, "value2": 5, "value3": 6.3},
    ]


def test_make_prediction_dataframe(
    analysis_data,
    patched_fero_client,
    prediction_dataframe,
    batch_prediction_success_data_cache,
):
    """Test that make prediction returns expected response as a dataframe."""
    analysis = Analysis(patched_fero_client, analysis_data)
    analysis._upload_file = mock.MagicMock()
    analysis._upload_file.return_value = "test_workspace_id"
    analysis._poll_workspace_for_prediction = mock.MagicMock()

    # Mock prediction
    test_prediction = Prediction(patched_fero_client, "result_id")
    test_prediction._data_cache = batch_prediction_success_data_cache
    test_prediction._complete = True
    test_prediction.get_results = mock.MagicMock()
    test_prediction.get_results.return_value = prediction_dataframe  # As pd.DataFrame
    analysis._poll_workspace_for_prediction.return_value = test_prediction

    results = analysis.make_prediction(prediction_dataframe)

    assert isinstance(results, pd.DataFrame)
    expected = pd.DataFrame(
        [
            {"value1": 1.0, "value2": 2, "value3": 3.3},
            {"value1": 4.0, "value2": 5, "value3": 6.3},
        ]
    )
    assert_frame_equal(
        results,
        expected,
        check_like=True,
    )


def test_make_prediction_prediction_failure(
    analysis_data,
    patched_fero_client,
    prediction_data,
    batch_prediction_failure_data_cache,
):
    """Test that a FeroError is raised if a prediction fails."""
    with pytest.raises(FeroError):
        analysis = Analysis(patched_fero_client, analysis_data)
        analysis._upload_file = mock.MagicMock()
        analysis._upload_file.return_value = "test_workspace_id"
        analysis._poll_workspace_for_prediction = mock.MagicMock()
        test_prediction = Prediction(patched_fero_client, "result_id")
        test_prediction._data_cache = batch_prediction_failure_data_cache
        test_prediction._complete = True
        analysis._poll_workspace_for_prediction.return_value = test_prediction

        analysis.make_prediction(prediction_data)


def test_analysis_factor_names(test_analysis_with_data):
    """Test that factor names are parsed correctly."""
    assert test_analysis_with_data.factor_names == [
        "Factor 1",
        "Factor 2",
        "Factor 3",
        "Category 0",
    ]


def test_analysis_target_names(test_analysis_with_data):
    """Test that factor names are parsed correctly."""
    assert test_analysis_with_data.target_names == ["Target 1", "Target 2"]


def test_make_optimization_fault_blueprint(test_analysis_with_data):
    """Test that a FeroError is raised if this is a fault blueprint."""
    test_analysis_with_data._data["blueprint_name"] = "fault"

    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test optimization",
            {
                "goal": "maximize",
                "factor": {"name": "Factor 2", "min": 70.0, "max": 100.0},
            },
            [
                {"name": "Factor 1", "min": 60.0, "max": 200.0},
                {"name": "Target 1", "min": 600.0, "max": 700.0},
            ],
        )


def test_make_optimization_goal_not_in_analysis(test_analysis_with_data):
    """Test that a FeroError is raised if the goal doesn't include columns in the analysis."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "goal": "maximize",
                "factor": {"name": "factor 4", "min": 5.0, "max": 7.0},
            },
            [{"name": "Target 1", "min": 10, "max": 11}],
        )


def test_make_optimization_no_constraints(test_analysis_with_data):
    """Test that a FeroError is raised if constraints are specified."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "goal": "maximize",
                "factor": {"name": "Target 1", "min": 5.0, "max": 7.0},
            },
            [],
        )


def test_make_optimization_bad_goals(test_analysis_with_data):
    """Test that a FeroError is raised if the goal field isn't valid."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "goal": "circumambulate",
                "factor": {"name": "Factor 1", "min": 5.0, "max": 7.0},
            },
            [{"name": "Target 1", "min": 10, "max": 11}],
        )


def test_make_optimization_wrong_format_goal_factors(test_analysis_with_data):
    """Test that a FeroError is raised if the factor is malformed."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "goal": "maximize",
                "factor": {"name": "Factor 1", "smallest": 5.0, "max": 7.0},
            },
            [{"name": "Target 1", "min": 10, "max": 11}],
        )


def test_make_optimization_type_not_cost(test_analysis_with_data):
    """Test that a FeroError is raised if the goal config is a type, but not cost."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "type": "chad",
                "goal": "maximize",
                "cost_function": [
                    {"name": "Factor 1", "min": 5.0, "max": 7.0, "cost": 1000}
                ],
            },
            [{"name": "Target 1", "min": 10, "max": 11}],
        )


def test_make_optimization_type_cost_no_cost_function(test_analysis_with_data):
    """Test that a FeroError is raised if the type is cost but there is no cost_function key."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "type": "COST",
                "goal": "maximize",
                "factor": {"name": "Factor 1", "smallest": 5.0, "max": 7.0},
            },
            [{"name": "Target 1", "min": 10, "max": 11}],
        )


def test_make_optimization_type_cost_malformed_cost_function(test_analysis_with_data):
    """Test that a type COST optimization raises a fero error if the cost functions are malformed."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "type": "COST",
                "goal": "maximize",
                "cost_function": [
                    {"name": "Factor 1", "min": 5.0, "max": 7.0, "price": 1000}
                ],
            },
            [{"name": "Target 1", "min": 10, "max": 11}],
        )


def test_make_optimization_type_cost_more_than_three(test_analysis_with_data):
    """Test that a type COST optimization raises a fero error if more than 3 cost functions are specified."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "type": "COST",
                "goal": "maximize",
                "cost_function": [
                    {"name": "Factor 1", "min": 5.0, "max": 7.0, "cost": 1000},
                    {"name": "Factor 2", "min": 5.0, "max": 7.0, "cost": 1000},
                    {"name": "Factor 3", "min": 5.0, "max": 7.0, "cost": 1000},
                    {"name": "Factor 4", "min": 5.0, "max": 7.0, "cost": 1000},
                ],
            },
            [{"name": "Target 1", "min": 10, "max": 11}],
        )


def test_make_optimization_type_cost_target_not_in_function_or_constraints(
    test_analysis_with_data,
):
    """Test that a type COST optimization raises a Fero Error if a target not in either the cost function or the constaints."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "type": "COST",
                "goal": "maximize",
                "cost_function": [
                    {"name": "Factor 1", "min": 5.0, "max": 7.0, "cost": 1000},
                ],
            },
            [{"name": "Factor 2", "min": 10, "max": 11}],
        )


def test_make_optimization_constraints_not_in_analysis(test_analysis_with_data):
    """Test that a FeroError is raised if the constraints are not in the analysis."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "goal": "maximize",
                "factor": {"name": "Factor 1", "min": 5.0, "max": 7.0},
            },
            [{"name": "Target 4", "min": 10, "max": 11}],
        )


def test_make_optimization_categorical_goal(test_analysis_with_data):
    """Test that a FeroError is raised if the goal is category."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "goal": "maximize",
                "factor": {"name": "Category 0", "min": 5.0, "max": 7.0},
            },
            [{"name": "Target 4", "min": 10, "max": 11}],
        )


def test_make_optimization_type_cost_categorical_function(test_analysis_with_data):
    """Test that a type COST optimization raises a Fero Error if a category is a cost funciton."""
    with pytest.raises(FeroError):
        test_analysis_with_data.make_optimization(
            "test",
            {
                "type": "COST",
                "goal": "maximize",
                "cost_function": [
                    {"name": "Category 0", "min": 5.0, "max": 7.0, "cost": 1000},
                ],
            },
            [{"name": "Target 1", "min": 10, "max": 11}],
        )


def test_analysis_make_optimization_simple_case(
    test_analysis_with_data, expected_optimization_config
):
    """Test that make_optimzation makes expected requests and returns a prediction."""
    pred = test_analysis_with_data.make_optimization(
        "test optimization",
        {
            "goal": "maximize",
            "factor": {"name": "Factor 2", "min": 70.0, "max": 100.0},
        },
        [
            {"name": "Factor 1", "min": 60.0, "max": 200.0},
            {"name": "Target 1", "min": 600.0, "max": 700.0},
        ],
        include_confidence_intervals=False,
    )

    assert pred.result_id == "f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7"
    test_analysis_with_data._client.post.assert_called_with(
        f"/api/analyses/{str(test_analysis_with_data.uuid)}/workspaces/",
        {"request": expected_optimization_config, "visible": False},
    )
    test_analysis_with_data._client.get.assert_called()


def test_analysis_make_optimization_simple_case_categorical(
    test_analysis_with_data, expected_optimization_config
):
    """Test that make_optimzation makes expected requests with a categorical value and returns a prediction."""
    pred = test_analysis_with_data.make_optimization(
        "test optimization",
        {
            "goal": "maximize",
            "factor": {"name": "Factor 2", "min": 70.0, "max": 100.0},
        },
        [
            {"name": "Category 0"},
            {"name": "Target 1", "min": 600.0, "max": 700.0},
        ],
        include_confidence_intervals=False,
    )
    expected_optimization_config["input_data"]["bounds"][0] = {
        "factor": "Category 0",
        "dtype": "factor_category",
    }
    assert pred.result_id == "f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7"
    test_analysis_with_data._client.post.assert_called_with(
        f"/api/analyses/{str(test_analysis_with_data.uuid)}/workspaces/",
        {"request": expected_optimization_config, "visible": False},
    )
    test_analysis_with_data._client.get.assert_called()


def test_analysis_make_optimization_cost(
    test_analysis_with_data, expected_optimization_config
):
    """Test that make_optimzation makes expected requests and prediction for a cost optimization."""
    expected_optimization_config["input_data"]["objectives"] = [
        {"factor": "FERO_COST_FUNCTION", "goal": "maximize"}
    ]

    expected_optimization_config["input_data"]["bounds"] = [
        {
            "factor": "Factor 1",
            "lowerBound": 60.0,
            "upperBound": 200.0,
            "dtype": "factor_float",
        },
        {
            "factor": "Target 1",
            "lowerBound": 600.0,
            "upperBound": 700.0,
            "dtype": "target_float",
            "confidenceInterval": "exclude",
        },
        {
            "factor": "Factor 2",
            "lowerBound": 70.0,
            "upperBound": 100.0,
            "dtype": "factor_float",
        },
        {
            "factor": "Factor 3",
            "lowerBound": 100.0,
            "upperBound": 150.0,
            "dtype": "factor_float",
        },
        {
            "factor": "FERO_COST_FUNCTION",
            "lowerBound": 2700.00,
            "upperBound": 4000.0,
            "dtype": "function",
        },
    ]

    expected_optimization_config["input_data"]["linearFunctionDefinitions"] = {
        "FERO_COST_FUNCTION": {
            "Factor 2": 10.0,
            "Factor 3": 20.0,
        }
    }

    pred = test_analysis_with_data.make_optimization(
        "test optimization",
        {
            "type": "COST",
            "goal": "maximize",
            "cost_function": [
                {"name": "Factor 2", "min": 70.0, "max": 100.0, "cost": 10.0},
                {"name": "Factor 3", "min": 100.0, "max": 150.0, "cost": 20.0},
            ],
        },
        [
            {"name": "Factor 1", "min": 60.0, "max": 200.0},
            {"name": "Target 1", "min": 600.0, "max": 700.0},
        ],
        include_confidence_intervals=False,
    )

    assert pred.result_id == "f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7"
    test_analysis_with_data._client.post.assert_called_with(
        f"/api/analyses/{str(test_analysis_with_data.uuid)}/workspaces/",
        {"request": expected_optimization_config, "visible": False},
    )


def test_analysis_make_optimization_simple_case_basis_override(
    test_analysis_with_data, expected_optimization_config
):
    """Test that make_optimzation makes expected requests and returns a prediction with basis_value overrides."""
    expected_optimization_config["input_data"]["basisValues"]["Factor 1"] = 150.0
    pred = test_analysis_with_data.make_optimization(
        "test optimization",
        {
            "goal": "maximize",
            "factor": {"name": "Factor 2", "min": 70.0, "max": 100.0},
        },
        [
            {"name": "Factor 1", "min": 60.0, "max": 200.0},
            {"name": "Target 1", "min": 600.0, "max": 700.0},
        ],
        fixed_factors={"Factor 1": 150.00},
        include_confidence_intervals=False,
    )

    assert pred.result_id == "f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7"
    test_analysis_with_data._client.post.assert_called_with(
        f"/api/analyses/{str(test_analysis_with_data.uuid)}/workspaces/",
        {"request": expected_optimization_config, "visible": False},
    )


def test_analysis_make_optimization_include_confidence(
    test_analysis_with_data, expected_optimization_config
):
    """Test that make_optimzation makes expected requests and returns a prediction with confidence intervals requested."""
    expected_optimization_config["input_data"]["bounds"][1][
        "confidenceInterval"
    ] = "include"
    pred = test_analysis_with_data.make_optimization(
        "test optimization",
        {
            "goal": "maximize",
            "factor": {"name": "Factor 2", "min": 70.0, "max": 100.0},
        },
        [
            {"name": "Factor 1", "min": 60.0, "max": 200.0},
            {"name": "Target 1", "min": 600.0, "max": 700.0},
        ],
        include_confidence_intervals=True,
    )

    assert pred.result_id == "f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7"
    test_analysis_with_data._client.post.assert_called_with(
        f"/api/analyses/{str(test_analysis_with_data.uuid)}/workspaces/",
        {"request": expected_optimization_config, "visible": False},
    )


def test_analysis_make_optimization_synchronous_false(
    test_analysis_with_data, expected_optimization_config
):
    """Test that make_optimzation makes expected requests and returns a prediction with synchronous false."""
    pred = test_analysis_with_data.make_optimization(
        "test optimization",
        {
            "goal": "maximize",
            "factor": {"name": "Factor 2", "min": 70.0, "max": 100.0},
        },
        [
            {"name": "Factor 1", "min": 60.0, "max": 200.0},
            {"name": "Target 1", "min": 600.0, "max": 700.0},
        ],
        include_confidence_intervals=False,
        synchronous=False,
    )

    assert pred.result_id == "f0123ab1-c6f4-4bd1-b1a6-02896ba57fc7"
    test_analysis_with_data._client.post.assert_called_with(
        f"/api/analyses/{str(test_analysis_with_data.uuid)}/workspaces/",
        {"request": expected_optimization_config, "visible": False},
    )
    test_analysis_with_data._client.get.assert_not_called()


def test_prediction_complete_false(
    patched_fero_client, prediction_results_response_started
):
    """Test that prediction.complete is false if not completed."""
    patched_fero_client.get.return_value = prediction_results_response_started

    pred = Prediction(patched_fero_client, "c9448486-0c59-487f-9c9c-345d98103fcb")
    assert pred.complete is False


def test_prediction_complete_true(
    patched_fero_client, prediction_results_response_completed
):
    """Test that prediction.complete is true if completed."""
    patched_fero_client.get.return_value = prediction_results_response_completed

    pred = Prediction(patched_fero_client, "c9448486-0c59-487f-9c9c-345d98103fcb")
    assert pred.complete is True


def test_prediction_get_result_not_complete(
    patched_fero_client, prediction_results_response_started
):
    """Test that a Fero Error is raised if get_result is called before a prediction is complete."""
    patched_fero_client.get.return_value = prediction_results_response_started
    pred = Prediction(patched_fero_client, "c9448486-0c59-487f-9c9c-345d98103fcb")

    with pytest.raises(FeroError) as err:
        pred.get_results()

    assert err.value.message == "Prediction is not complete."


def test_prediction_get_result_dataframe(
    patched_fero_client, prediction_results_response_completed
):
    """Test that the expected data frame is generated by get_results."""
    patched_fero_client.get.return_value = prediction_results_response_completed
    pred = Prediction(patched_fero_client, "c9448486-0c59-487f-9c9c-345d98103fcb")

    excepted = pd.DataFrame(
        [
            [
                148.02718539320762,
                90.234,
                148.02718539320762,
                666.66666,
                677.1234,
                699.1234,
                700.4567,
                677.1234,
                699.1234,
            ],
        ],
        index=[0],
        columns=[
            "Factor 1",
            "Factor 2",
            "Factor 3",
            "Target 1 (5%)",
            "Target 1 (Mean)",
            "Target 1 (95%)",
            "Target 2 (5%)",
            "Target 2 (Mean)",
            "Target 2 (95%)",
        ],
    )

    assert_frame_equal(excepted, pred.get_results())


def test_prediction_get_result_dict(
    patched_fero_client, prediction_results_response_completed
):
    """Test that the expected list of dicts is returned for format records."""
    patched_fero_client.get.return_value = prediction_results_response_completed
    pred = Prediction(patched_fero_client, "c9448486-0c59-487f-9c9c-345d98103fcb")

    excepted = [
        {
            "Factor 1": 148.02718539320762,
            "Factor 2": 90.234,
            "Factor 3": 148.02718539320762,
            "Target 1 (5%)": 666.66666,
            "Target 1 (Mean)": 677.1234,
            "Target 1 (95%)": 699.1234,
            "Target 2 (5%)": 700.4567,
            "Target 2 (Mean)": 677.1234,
            "Target 2 (95%)": 699.1234,
        }
    ]
    assert excepted == pred.get_results(format="records")


def test_analysis_upload_file_makes_expected_calls_azure(
    analysis_data, patched_fero_client
):
    """Test that _upload_file makes the expected calls and returns the expected values."""
    patched_fero_client.post.return_value = {
        "upload_type": "azure",
        "workspace_id": "uuid",
        "other_info": "test_value",
    }
    analysis = Analysis(patched_fero_client, analysis_data)
    analysis._client._azure_upload = mock.MagicMock()
    analysis._client._s3_upload = mock.MagicMock()
    fp = io.StringIO("test_data")
    analysis._upload_file(fp, "test_tag", "test_type")

    patched_fero_client.post.assert_has_calls(
        [
            mock.call(
                f"/api/analyses/{str(analysis.uuid)}/workspaces/inbox_url/",
                {"file_tag": "test_tag", "prediction_type": "test_type"},
            )
        ]
    )
    analysis._client._azure_upload.assert_has_calls(
        [mock.call({"upload_type": "azure", "other_info": "test_value"}, fp)]
    )
    analysis._client._s3_upload.assert_has_calls([])


def test_analysis_upload_file_makes_expected_calls_s3(
    analysis_data, patched_fero_client
):
    """Test that _upload_file makes the expected calls and returns the expected values."""
    patched_fero_client.post.return_value = {
        "upload_type": "s3",
        "workspace_id": "uuid",
        "other_info": "test_value",
    }
    analysis = Analysis(patched_fero_client, analysis_data)
    analysis._client._azure_upload = mock.MagicMock()
    analysis._client._s3_upload = mock.MagicMock()
    fp = io.StringIO("test_data")
    analysis._upload_file(fp, "test_tag", "test_type")

    patched_fero_client.post.assert_has_calls(
        [
            mock.call(
                f"/api/analyses/{str(analysis.uuid)}/workspaces/inbox_url/",
                {"file_tag": "test_tag", "prediction_type": "test_type"},
            )
        ]
    )
    analysis._client._s3_upload.assert_has_calls(
        [mock.call({"upload_type": "s3", "other_info": "test_value"}, "test_tag", fp)]
    )
    analysis._client._azure_upload.assert_has_calls([])


def test_is_retraining_not_retraining(analysis_data, patched_fero_client):
    """Test that is_retraining() returns false if the analysis is not training."""
    patched_fero_client.get.return_value = analysis_data.copy()
    analysis_data["latest_revision_model_state"] = "T"
    analysis = Analysis(patched_fero_client, analysis_data)
    assert analysis.is_retraining() is False
    patched_fero_client.get.assert_called_with(
        f"/api/analyses/{analysis_data['uuid']}/"
    )


def test_is_retraining_retraining(analysis_data, patched_fero_client):
    """Test that is_retraining() returns true if the analysis is training."""
    analysis_data["latest_revision_model_state"] = "T"
    patched_fero_client.get.return_value = analysis_data.copy()

    analysis = Analysis(patched_fero_client, analysis_data)
    assert analysis.is_retraining() is True
    patched_fero_client.get.assert_called_with(
        f"/api/analyses/{analysis_data['uuid']}/"
    )


def test_not_retraining_new_version_race_condition(analysis_data, patched_fero_client):
    """Test that is_retraining is called again if the latest version."""
    second_version = analysis_data.copy()
    second_version["latest_revision"] = second_version["latest_revision"] + 1

    patched_fero_client.get.side_effect = [
        analysis_data,
        second_version,
        second_version,
        second_version,
    ]

    analysis = Analysis(patched_fero_client, analysis_data)
    assert analysis.is_retraining() is False
    patched_fero_client.get.assert_has_calls(
        [mock.call(f"/api/analyses/{analysis_data['uuid']}/") for _ in range(4)]
    )


def test_revise_makes_expected_calls(analysis_data, patched_fero_client, revision_data):
    """Test that revise makes the expected calls to the server."""
    analysis = Analysis(patched_fero_client, analysis_data)
    patched_fero_client.get.side_effect = [
        analysis_data,
        analysis_data,
        revision_data,
        analysis_data,
        analysis_data,
    ]
    patched_fero_client.post.return_value = revision_data

    with mock.patch("datetime.datetime") as dt:
        now = datetime.datetime(2021, 8, 4, 0, 0, 1)
        dt.now.return_value = now
        analysis.revise()

        patched_fero_client.get.assert_has_calls(
            [
                mock.call(f"/api/analyses/{analysis_data['uuid']}/"),
                mock.call(
                    f"/api/analyses/{analysis_data['uuid']}/revisions/{analysis_data['latest_revision']}/"
                ),
                mock.call(f"/api/analyses/{analysis_data['uuid']}/"),
            ]
        )
        patched_fero_client.post.assert_called_with(
            f"/api/analyses/{analysis_data['uuid']}/revisions/",
            {
                "description": f"Revision created via fero client on {now.isoformat()}",
                "configured_blueprint": revision_data["configured_blueprint"],
            },
        )


def test_revise_does_not_revise_if_training(analysis_data, patched_fero_client):
    """Test a revision is not made if the analysis is currently being trained."""
    analysis = Analysis(patched_fero_client, analysis_data)
    analysis_data["latest_revision_model_state"] = "T"
    patched_fero_client.get.return_value = analysis_data

    analysis.revise()
    patched_fero_client.get.called_once_with(f"/api/analyses/{analysis_data['uuid']}/")
    patched_fero_client.post.assert_not_called()



"""A module to test the `Asset` class."""

import pytest
import json
from fero.asset import Asset
import pandas as pd
from pandas.testing import assert_frame_equal


@pytest.fixture
def test_asset(asset_data, patched_fero_client, prediction_result_success_data):
    """Get a sample `Asset` object."""
    patched_fero_client.post.return_value = prediction_result_success_data
    return Asset(patched_fero_client, asset_data)


@pytest.fixture
def default_predictions():
    """Get sample dataframe of asset predictions."""
    return pd.DataFrame(
        **{
            "data": [
                [
                    7.93735099382404,
                    7.253059448467135,
                    7.688163173887008,
                    8.201992282568096,
                    8.562654808412713,
                    7.290020590899204,
                    7.137018988443888,
                    7.222864154749371,
                    7.3457978216120265,
                    7.474805839163134,
                    5.0696783341277625,
                    3.6869059267194997,
                    4.470695153741833,
                    5.683819078338525,
                    6.529167263822492,
                    1.7271781080698847,
                    1.2293876958841274,
                    1.5332391289059648,
                    1.9211059805503128,
                    2.1972651616834313,
                ],
                [
                    8.059697064287855,
                    6.962005033013789,
                    7.721065781687776,
                    8.453028511724858,
                    8.970620656521195,
                    7.385830908631496,
                    7.2164206074582635,
                    7.312782100772867,
                    7.4474424849045455,
                    7.603513646484785,
                    4.969889551638316,
                    3.2410276157675737,
                    4.180035194039614,
                    5.742825222050875,
                    6.794145544189322,
                    1.6465033404595186,
                    1.0098659780967365,
                    1.397920672348523,
                    1.924213021790375,
                    2.2023476424790025,
                ],
                [
                    8.192820567540917,
                    6.753720344190171,
                    7.691716734161963,
                    8.67238696189184,
                    9.56488526360722,
                    7.44409425873223,
                    7.2532241117672855,
                    7.3604140870516,
                    7.502760719158076,
                    7.7434025244433995,
                    4.854845925095373,
                    2.641086035815071,
                    3.913470878543495,
                    5.7220130262037365,
                    7.075994865914881,
                    1.5516005049741246,
                    0.7784695291543537,
                    1.2338449666203226,
                    1.8708144728173635,
                    2.318108568235832,
                ],
                [
                    8.349753362887295,
                    6.552024567348383,
                    7.619240288083308,
                    8.990032839229166,
                    10.303746588561847,
                    7.460260370638008,
                    7.250522528886993,
                    7.369555018121625,
                    7.511282423365672,
                    7.813681019224355,
                    4.6152964634338804,
                    1.934912207059016,
                    3.491997640608437,
                    5.690586705047334,
                    7.333590667213361,
                    1.416888407225492,
                    0.45107889112482186,
                    1.0140468362241668,
                    1.8300442319324934,
                    2.375700217914943,
                ],
                [
                    8.492439329004604,
                    6.199366100810615,
                    7.498189870344907,
                    9.30211022900819,
                    11.131550725647848,
                    7.482777901396371,
                    7.244147925303968,
                    7.3802383232561874,
                    7.530010225868999,
                    7.9247913725868155,
                    4.379943452590086,
                    1.282381711049496,
                    3.115976079652962,
                    5.591919257727733,
                    7.400851518682516,
                    1.2809534054200706,
                    0.1468654105352288,
                    0.802657230404595,
                    1.76232849577855,
                    2.424137436651667,
                ],
            ],
            "index": [
                "2019-04-29T20:00:00Z",
                "2019-04-29T21:00:00Z",
                "2019-04-29T22:00:00Z",
                "2019-04-29T23:00:00Z",
                "2019-04-30T00:00:00Z",
            ],
            "columns": [
                "mean:Factor 4",
                "p5:Factor 4",
                "p25:Factor 4",
                "p75:Factor 4",
                "p95:Factor 4",
                "mean:Factor 5",
                "p5:Factor 5",
                "p25:Factor 5",
                "p75:Factor 5",
                "p95:Factor 5",
                "mean:Factor 9",
                "p5:Factor 9",
                "p25:Factor 9",
                "p75:Factor 9",
                "p95:Factor 9",
                "mean:Asset 1",
                "p5:Asset 1",
                "p25:Asset 1",
                "p75:Asset 1",
                "p95:Asset 1",
            ],
        }
    )


@pytest.fixture
def test_asset_with_data(test_asset, default_predictions):
    """Get a sample `Asset` with loaded Scenario Simulator predictions."""
    test_asset._presentation_data_cache = [
        {
            "id": "sensor_forecaster",
            "tab": "Scenario Simulator",
            "type": "sensor_forecaster",
            "title": "Scenario Simulator",
            "content": {
                "factors": [
                    {"name": "Factor 4"},
                    {"name": "Factor 5"},
                    {"name": "Factor 9"},
                ],
                "sensors": [{"name": "Factor 3"}, {"name": "Factor 6"}],
                "targets": [{"name": "Asset 1"}],
                "default_predictions": default_predictions.to_dict(orient="split"),
            },
        }
    ]

    return test_asset


@pytest.fixture
def prediction_result_success_data():
    """Get sample data for a successful asset prediction."""
    return {
        "status": "SUCCESS",
        "data": {
            "index": [
                "2019-04-29T20:00:00Z",
                "2019-04-29T21:00:00Z",
                "2019-04-29T22:00:00Z",
                "2019-04-29T23:00:00Z",
                "2019-04-30T00:00:00Z",
            ],
            "columns": [
                "mean:Asset 1",
                "p5:Asset 1",
                "p25:Asset 1",
                "p75:Asset 1",
                "p95:Asset 1",
            ],
            "data": [
                [3.0, 4.0, 5.0, 6.0, 7.0],
                [1.0, 2.0, 3.0, 4.0, 5.0],
                [2.0, 3.0, 4.0, 5.0, 6.0],
                [4.0, 5.0, 6.0, 7.0, 8.0],
                [5.0, 6.0, 7.0, 8.0, 9.0],
            ],
        },
    }


def test_creates_asset_correctly(asset_data, patched_fero_client):
    """Test that a valid asset is created from valid data and a valid client."""
    asset = Asset(patched_fero_client, asset_data)

    assert isinstance(asset, Asset)
    assert str(asset) == "<Asset name=TEST ASSET>"


def test_has_trained_model_true(asset_data, patched_fero_client):
    """Test that `has_trained_model` is true if there is a latest configuration model."""
    asset = Asset(patched_fero_client, asset_data)
    assert asset.has_trained_model() is True


def test_has_trained_model_false(asset_data, patched_fero_client):
    """Test that `has_trained_model` is false if there is no configuration model."""
    asset_data["latest_completed_model"] = None
    asset = Asset(patched_fero_client, asset_data)
    assert asset.has_trained_model() is False


def test_predict_default(test_asset_with_data, default_predictions):
    """Test that default the prediction gives the expected result."""
    result = test_asset_with_data.predict()
    assert_frame_equal(result, default_predictions)


def test_predict_specified_factor_df(
    test_asset_with_data, default_predictions, prediction_result_success_data
):
    """Test a prediction with a manually specified `DataFrame` of factors."""
    specified = pd.DataFrame(
        {"Factor 4": default_predictions["mean:Factor 4"].to_list()}
    )
    result = test_asset_with_data.predict(specified)
    success_data = pd.DataFrame(**prediction_result_success_data["data"])
    expected_frame = default_predictions.drop(
        [c for c in default_predictions.columns if "Factor 4" in c], axis=1
    )
    expected_frame["specified:Factor 4"] = default_predictions["mean:Factor 4"]
    target_columns = [
        "mean:Asset 1",
        "p5:Asset 1",
        "p25:Asset 1",
        "p75:Asset 1",
        "p95:Asset 1",
    ]
    for c in target_columns:
        expected_frame[c] = success_data[c]
    assert_frame_equal(expected_frame, result)


def test_predict_specified_factor_dict(
    test_asset_with_data, default_predictions, prediction_result_success_data
):
    """Test a prediction with a manually specified `dict` of factors."""
    specified = {"Factor 4": default_predictions["mean:Factor 4"].to_list()}
    result = test_asset_with_data.predict(specified)
    success_df = pd.DataFrame(**prediction_result_success_data["data"])
    expected_dict = default_predictions.drop(
        [c for c in default_predictions.columns if "Factor 4" in c], axis=1
    ).to_dict("list")
    expected_dict["index"] = success_df.index.to_list()
    expected_dict["specified:Factor 4"] = default_predictions["mean:Factor 4"].to_list()
    target_columns = [
        "mean:Asset 1",
        "p5:Asset 1",
        "p25:Asset 1",
        "p75:Asset 1",
        "p95:Asset 1",
    ]
    for c in target_columns:
        expected_dict[c] = success_df[c].to_list()
    assert json.dumps(expected_dict, sort_keys=True) == json.dumps(
        result, sort_keys=True
    )
